from redis import Redis, ConnectionPool

from fred.dao.service.utils import get_redis_configs_from_payload
from fred.dao.service.interface import ServiceConnectionPoolInterface, ServiceInterface


class RedisConnectionPool(ServiceConnectionPoolInterface[ConnectionPool]):

    @classmethod
    def _create_pool(cls, **kwargs) -> ConnectionPool:
        configs = get_redis_configs_from_payload(payload=kwargs, keep=False)
        return ConnectionPool(**configs)


class RedisService(ServiceInterface[Redis]):
    instance: Redis

    @classmethod
    def _create_instance(cls, **kwargs) -> Redis:
        return Redis(connection_pool=RedisConnectionPool.get_or_create_pool(**kwargs))

    @classmethod
    def auto(cls, **kwargs) -> "RedisService":
        cls.instance = Redis(connection_pool=RedisConnectionPool.get_or_create_pool(**kwargs))
        return cls(**kwargs)



from dataclasses import dataclass
from threading import Thread
from typing import Optional

from fred.future import Future
from fred.settings import logger_manager
from fred.utils.dateops import datetime_utcnow
from fred.worker.runner.status import RunnerStatus
from fred.worker.runner.model._runner_spec import RunnerSpec
from fred.worker.runner.backend import RunnerBackend
from fred.worker.runner.settings import FRD_RUNNER_BACKEND


logger = logger_manager.get_logger(name=__name__)


@dataclass(frozen=True, slots=True)
class PluginExecutionOutput:
    runner_id: str
    future_exec: Future
    future_monitor: Optional[Future] = None


@dataclass(frozen=True, slots=True)
class PluginInterface:
    backend: RunnerBackend

    @classmethod
    def auto(cls, service_name: Optional[str] = None, **kwargs) -> "PluginInterface":
        """Auto-instantiate the plugin interface with backend services.
        If a backend service is not provided, it will create one using configurations
        extracted from the provided keyword arguments.
        Args:
            **kwargs: Keyword arguments that may contain backend service configurations or an existing backend instance.
        Returns:
            PluginInterface: An instance of the PluginInterface with backend services.
        """
        backend = RunnerBackend.auto(
            service_name=service_name or FRD_RUNNER_BACKEND,
            **kwargs
        )
        return cls(backend=backend)
    
    def _execute(
            self,
            spec: RunnerSpec,
            **kwargs
        ):
        raise NotImplementedError("This method should be overridden by subclasses.")

    def _monitor(self, spec: RunnerSpec, **kwargs):
        raise NotImplementedError("This method should be overridden by subclasses.")

    def _execute_wrapper(
            self,
            spec: RunnerSpec,
            **kwargs
        ) -> str:
        """Wrapper method to handle execution and include error logging.

        Since we don't control the implementation of the _execute method in subclasses, we
        need to wrap it to add error handling and logging.

        Args:
            runner_info (RunnerInfo): Information about the runner to execute.
            outer_handler (RunnerHandler): The outer handler to use for execution.
            **kwargs: Additional keyword arguments to pass to the execution method implemented by the subclass.
        """
        runner_id = spec.runner_id
        runner_status = self.backend.keyval(
            key=RunnerStatus.get_key(runner_id=runner_id)
        )
        try:
            self._execute(spec=spec, **kwargs)
        except Exception as e:
            runner_status.set(RunnerStatus.ERROR.get_val(str(e)))
            logger.error(f"Error executing runner '{runner_id}': {e}")
            raise
        return runner_id
    
    def execute(
            self,
            spec: RunnerSpec,
            wait_for_exec: bool = False,
            wait_for_monitor: bool = False,
            timeout: Optional[int] = None,
            enable_monitor: bool = False,
            **kwargs
        ) -> PluginExecutionOutput:
        runner_id = spec.runner_id
        logger.info(f"Starting thread runner '{runner_id}' using plugin '{self.__class__.__name__}'.")
        future_exec = Future(
            function=self._execute_wrapper,
            spec=spec,
            **kwargs
        )
        future_monitor = None
        if enable_monitor:
            future_monitor = self.monitor(spec=spec, **kwargs)
        else:
            logger.info(f"Monitoring disabled for runner '{runner_id}'.")
        if wait_for_exec:
            logger.info(f"Waiting for execution of runner '{runner_id}' to complete.")
            future_exec.wait(timeout=timeout)
        if future_monitor and wait_for_monitor:
            future_monitor.wait(timeout=timeout)
        return PluginExecutionOutput(
            runner_id=runner_id,
            future_exec=future_exec,
            future_monitor=future_monitor,
        )
    
    def monitor(
            self,
            spec: RunnerSpec,
            **kwargs
        ) -> Optional[Thread]:
        """Start monitoring the runner in a separate thread.
        Args:
            runner_info (RunnerInfo): Information about the runner to monitor.
            blocking (bool, optional): Whether to block the main thread until monitoring is complete. Defaults to False.
            timeout (Optional[int], optional): Timeout in seconds for the monitoring thread. Defaults to None.
            **kwargs: Additional keyword arguments to pass to the monitoring method.
        Returns:
            Optional[Thread]: The thread object for the monitoring thread if not blocking, else None."""
        # Start monitoring in a separate thread
        return Future(
            function=self._monitor,
            spec=spec,
            **kwargs
        )



from fred.settings import get_environ_variable


FRD_RUNNER_BACKEND = get_environ_variable(
    name="FRD_RUNNER_BACKEND",
    default="STDLIB",
).upper()

FRD_RUNNER_REQUEST_QUEUE = get_environ_variable(
    name="FRD_RUNNER_REQUEST_QUEUE",
    default=None,
)

FRD_RUNNER_RESPONSE_QUEUE = get_environ_variable(
    name="FRD_RUNNER_RESPONSE_QUEUE",
    default=None,
)



from dataclasses import dataclass

from fred.settings import logger_manager
from fred.worker.runner.info import RunnerInfo
from fred.worker.runner.handler import RunnerHandler
from fred.worker.runner.model._runner_spec import RunnerSpec
from fred.worker.runner.plugins.interface import PluginInterface

logger = logger_manager.get_logger(name=__name__)


@dataclass(frozen=True, slots=True)
class LocalPlugin(PluginInterface):

    def _execute(
            self,
            spec: RunnerSpec,
            **kwargs
        ):
        """Execute the runner locally by directly invoking the handler's run method.
        This method bypasses any queuing mechanism and runs the handler in the current process.
        Args:
            spec (RunnerSpec): The specification of the runner to execute.
            **kwargs: Additional keyword arguments that may be used for execution.
        """
        outer_handler_classname = kwargs.pop("outer_handler_classname", "RunnerHandler")
        outer_handler_classpath = kwargs.pop("outer_handler_classpath", "fred.worker.runner.handler")
        outer_handler_init_kwargs = kwargs.pop("outer_handler_init_kwargs", {})
        outer_handler = RunnerHandler.find_handler(
            handler_classname=outer_handler_classname,
            import_pattern=outer_handler_classpath,
            **outer_handler_init_kwargs,
        )
        outer_handler.run(event=spec.as_event(), as_future=False)



from dataclasses import dataclass
from typing import Optional

from fred.settings import logger_manager
from fred.dao.service.catalog import ServiceCatalog
from fred.dao.comp.interface import ComponentInterface

logger = logger_manager.get_logger(name=__name__)


@dataclass(frozen=True, slots=True)
class FredKeyVal(ComponentInterface):
    """A simple key-value store implementation using a backend service.
    This class provides methods to interact with a key-value store, such as setting,
    getting, and deleting key-value pairs. The actual implementation of these methods
    depends on the underlying service being used (e.g., Redis).
    """
    key: str
    
    def set(self, value: str, key: Optional[str] = None, **kwargs) -> None:
        """Sets a key-value pair in the store.
        The implementation of this method depends on the underlying service.
        For example, if the service is Redis, it uses the SET command to store the
        key-value pair.
        Args:
            key (str): The key to set.
            value (str): The value to associate with the key.
            **kwargs: Additional keyword arguments for setting the key-value pair,
                      such as expiration time.
        Raises:
            NotImplementedError: If the method is not implemented for the current service.
        """
        key = key or self.key
        match self._cat:
            case ServiceCatalog.REDIS:
                self._srv.client.set(key, value)
                expire = kwargs.get("expire")
                if expire and isinstance(expire, int):
                    self._srv.client.expire(key, expire)
            case ServiceCatalog.STDLIB:
                self._srv.client._memstore_keyval[key] = value
                # TODO: Implement expiration logic
                if "expire" in kwargs:
                    logger.warning("Expiration not implemented for STDLIB service.")
            case _:
                raise NotImplementedError(f"Set method not implemented for service {self._nme}")

    def get(self, key: Optional[str] = None, fail: bool = False) -> Optional[str]:
        """Gets the value associated with a key from the store.
        The implementation of this method depends on the underlying service.
        For example, if the service is Redis, it uses the GET command to retrieve the
        value associated with the key.
        Args:
            key (str): The key to retrieve.
            fail (bool): If True, raises a KeyError if the key is not found. Defaults to False.
        Returns:
            Optional[str]: The value associated with the key, or None if the key is not found
                           and fail is False.
        Raises:
            KeyError: If the key is not found and fail is True.
            NotImplementedError: If the method is not implemented for the current service.
        """
        key = key or self.key
        result = None
        match self._cat:
            case ServiceCatalog.REDIS:
                result = self._srv.client.get(key)
            case ServiceCatalog.STDLIB:
                result = self._srv.client._memstore_keyval.get(key)
            case _:
                raise NotImplementedError(f"Get method not implemented for service {self._nme}")
        if fail and result is None:
            raise KeyError(f"Key {key} not found.")
        return result

    def delete(self, key: Optional[str] = None) -> None:
        """Deletes a key-value pair from the store.
        The implementation of this method depends on the underlying service.
        For example, if the service is Redis, it uses the DEL command to remove the
        key-value pair.
        Args:
            key (str): The key to delete.
        Raises:
            NotImplementedError: If the method is not implemented for the current service.
        """
        key = key or self.key
        match self._cat:
            case ServiceCatalog.REDIS:
                self._srv.client.delete(key)
            case ServiceCatalog.STDLIB:
                self._srv.client._memstore_keyval.pop(key, None)
            case _:
                raise NotImplementedError(f"Delete method not implemented for service {self._nme}")


import enum

from fred.future.callback.interface import CallbackInterface
from fred.future.callback._function import CallbackFunction



class CallbackCatalog(enum.Enum):
    FUNCTION = CallbackFunction

    def __call__(self, *args, **kwargs) -> CallbackInterface:
        return self.value(*args, **kwargs)



from fred.version import version
from fred.settings import logger_manager
from fred.cli.interface import AbstractCLI


logger = logger_manager.get_logger(name=__name__)


class CLIExtensionGroups:
    """CLI Extensions providing access to various integrations by following a lazy loading pattern."""

    @property
    def databricks(self):
        from fred.integrations.databricks.cli_ext import DatabricksExt
        return DatabricksExt()
    
    @property
    def runpod(self):
        from fred.integrations.runpod.cli_ext import RunPodExt
        return RunPodExt()
    
    @property
    def runner_server(self):
        from fred.worker.runner.rest.cli_ext import RunnerServerExt
        return RunnerServerExt()


class CLI(AbstractCLI, CLIExtensionGroups):

    def version(self) -> str:
        return version.value



from dataclasses import dataclass

from fred.maturity import Maturity, MaturityLevel
from fred.worker.interface import HandlerInterface
from fred.settings import logger_manager

logger = logger_manager.get_logger(name=__name__)


module_maturity = Maturity(
    level=MaturityLevel.TO_BE_DEPRECATED,
    reference=__name__,
    message=(
        "Runpod Helper implementation is soon to be deprecated! "
        "Please use the Fred-Worker interface instead (HandlerInterface)."
    )
)


@dataclass(frozen=True, slots=False)
class HandlerHelper(HandlerInterface):
    
    def __post_init__(self):
        super().__post_init__()
        logger.warning("Runpod Helper is soon to be deprecated! Please use Fred-Worker interface instead.")



import os
from typing import Dict, List, Optional
from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class Version:
    name: str
    value: str

    def get_upcoming_options(self) -> List['Version']:
        vals = self.components(as_int=True)
        return [
            Version(
                name=self.name,
                value=".".join(
                    str(v)
                    for j, v in enumerate(vals[:i] + [val+1] + (2-i)*[0])
                ),
            )
            for i, val in enumerate(vals)
        ]

    def get_upcoming_options_with_tag(self) -> Dict[str, Dict]:
        tags = ["major", "minor", "patch"]
        return {
            version.value: {
                "version": version,
                "scope": tags[index],
            }
            for index, version in enumerate(self.get_upcoming_options())
        }

    def upcoming(
            self,
            major: bool = False,
            minor: bool = False,
            patch: bool = False,
    ) -> 'Version':
        ver_maj, ver_min, ver_pat = self.get_upcoming_options()
        if sum([major, minor, patch]) != 1:
            raise ValueError
        return ver_maj if major else ver_min if minor else ver_pat

    def components(self, as_int: bool = False) -> list:
        return [int(val) if as_int else val for val in self.value.split(".")]

    @property
    def major(self) -> int:
        component, *_ = self.components(as_int=True)
        return component

    @property
    def minor(self) -> int:
        _, component, *_ = self.components(as_int=True)
        return component

    @property
    def patch(self) -> int:
        *_, component = self.components(as_int=True)
        return component

    @classmethod
    def from_path(cls, dirpath: str, name: str):
        for file in os.listdir(dirpath):
            if file.lower().endswith("version"):
                filepath = os.path.join(dirpath, file)
                break
        else:
            raise ValueError("Version file not found for package name: " + name)

        with open(filepath, "r") as version_file:
            version_value = version_file.readline().strip()  # TODO: Validate version pattern via regex
            return cls(name=name, value=version_value)


version = Version.from_path(name="fred", dirpath=os.path.dirname(__file__))



from dataclasses import dataclass
from functools import wraps
from typing import (
    Callable,
    Generic,
    Generator,
    TypeVar,
)

from fred.settings import logger_manager
from fred.monad.interface import MonadInterface

logger = logger_manager.get_logger(__name__)

A = TypeVar("A")
B = TypeVar("B")


@dataclass(frozen=True, slots=True)
class EitherWrapper(Generic[A]):
    function: Callable[..., A]

    def __call__(self, *args, **kwargs) -> A:
        return self.function(*args, **kwargs)

    def either(self, *args, **kwargs)-> 'Either[A]':
        try:
            return Right[A].from_value(val=self.function(*args, **kwargs))
        except Exception as e:
            return Left(exception=e)


class Either(MonadInterface[A]):

    @classmethod
    def comprehension(cls, generator: Generator):
        try:
            return Either.from_value(val=next(generator))
        except Exception as e:
            return Left(exception=e)

    @classmethod
    def decorator(cls, function: Callable):
        return wraps(function, updated=())(EitherWrapper(function=function))

    @classmethod
    def from_value(cls, val: A) -> 'Either[A]':
        if isinstance(val, Exception):
            return Left(exception=val)
        return Right(value=val)

    def resolve(self) -> A:
        match self:
            case Right(value=value):
                return value
            case Left(exception=exception):
                raise exception
            case _:
                raise TypeError("Unknown Either type")
            
    def __iter__(self) -> Generator[A, None, None]:
        match self:
            case Right(value=value):
                yield from [value]
            case Left(exception=exception):
                logger.warning(f"Attempted to iterate over Left: {exception}")
                yield from []
            case _:
                raise TypeError("Unknown Either type")


@dataclass(frozen=True, slots=True)
class Right(Either[A]):
    value: A

    def flat_map(self, function: Callable[[A], Either[B]]) -> Either[B]:
        try:
            return function(self.value)
        except Exception as e:
            logger.exception(f"Error in flat_map: {e}")
            # Consider that a right that fails is a left
            return Left(exception=e)


@dataclass(frozen=True, slots=True)
class Left(Either[A]):
    exception: Exception

    def flat_map(self, function: Callable[[A], Either[B]]) -> Either[B]:
        return Left(exception=self.exception)  # Propagate exception
    
    def map(self, function: Callable[[A], B]) -> Either[B]:
        return Left(exception=self.exception)  # Propagate exception


class EitherMonad(Either[A]):
    Either = Either
    Left = Left
    Right = Right



import json
import uuid
from typing import Optional

from pydantic.dataclasses import dataclass

from fred.utils.dateops import datetime_utcnow
from fred.worker.runner.model._request import RunnerRequest
from fred.worker.runner.model.interface import ModelInterface


@dataclass(frozen=True, slots=True)
class RunnerItem(ModelInterface):
    item_id: str
    item_created_at: str
    item_payload: dict

    @classmethod
    def uuid(cls, payload: dict, uuid_hash: bool = False) -> "RunnerItem":
        item_id = payload.get("item_id") or (
            str(uuid.uuid5(uuid.NAMESPACE_DNS, json.dumps(payload, default=str)))
            if uuid_hash else str(uuid.uuid4())
        )
        return cls(
            item_id=item_id,
            item_created_at=datetime_utcnow().isoformat(),
            item_payload=payload
        )
    
    def as_payload(self) -> dict:
        return {
            "item_id": self.item_id,
            "item_created_at": self.item_created_at,
            **self.item_payload,
        }
    
    def as_request(
            self,
            use_hash: bool = False,
            request_id: Optional[str] = None
    ) -> RunnerRequest:
        return RunnerRequest.uuid(
            payload=self.as_payload(),
            request_id=request_id,
            uuid_hash=use_hash,
        )



import os
from enum import StrEnum, auto
from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class DatabricksRuntime:
    python_version: str
    databricks_runtime: str
    libraries: list[dict]

    @classmethod
    def from_catalog(cls, catalog: 'DatabricksRuntimeCatalog') -> 'DatabricksRuntime':
        config = catalog.get_configuration()
        return cls(
            python_version=config.get("python_version", "3.12.3"),
            databricks_runtime=config.get("databricks_runtime", "16.4 LTS"),
            libraries=config.get("libraries", []),
        )


class DatabricksRuntimeCatalog(StrEnum):
    LTS_16_4 = auto()

    @property
    def key(self) -> str:
        runtime_type, runtime_num = self.name.split("_", 1)
        return f"{runtime_num.replace('_', '.')} {runtime_type}"

    @property
    def url(self) -> str:
        import posixpath
        return posixpath.join(
            "https://docs.databricks.com",
            "aws/en/release-notes/runtime",
            self.key.replace(" ", "").replace("_", "").lower(),
        )

    @classmethod
    def from_string(cls, name: str) -> 'DatabricksRuntimeCatalog':
        formatted_name = name.replace(" ", "").replace(".", "_").upper()
        return cls[formatted_name]

    def get_filepath(self) -> str:
        filename = f"{self.key.replace(' ', '').replace('.', '_')}.json"
        return os.path.join(
            os.path.dirname(__file__),
            "runtimes",
            filename,
        )
    
    def get_configuration(self) -> dict:
        import json

        with open(self.get_filepath(), "r") as file:
            return json.load(file)
        
    def get_runtime(self) -> DatabricksRuntime:
        return DatabricksRuntime.from_catalog(self)



from dataclasses import dataclass

from fastapi import FastAPI

from fred.settings import logger_manager, get_environ_variable
from fred.worker.runner.rest.routers.catalog import RouterCatalog

logger = logger_manager.get_logger(name=__name__)


@dataclass(frozen=True, slots=True)
class RunnerServer:
    app: FastAPI
    include_routers: list[str]
    exclude_routers: list[str]

    @classmethod
    def auto(cls, **kwargs) -> "RunnerServer":
        # Include routers by checking on keyword argument or environment variable
        include_routers = kwargs.pop("include_routers", None) or get_environ_variable(
            "FRD_RUNNER_API_INCLUDE_ROUTERS",
            default=""
        )
        if isinstance(include_routers, str):
            include_routers = [
                name.upper()
                for router in include_routers.split(";")
                if (name := router.strip())
            ]
        # Exclude routers by checking on keyword argument or environment variable
        exclude_routers = kwargs.pop("exclude_routers", None) or get_environ_variable(
            "FRD_RUNNER_API_EXCLUDE_ROUTERS",
            default=""
        )
        if isinstance(exclude_routers, str):
            exclude_routers = [
                name.upper()
                for router in exclude_routers.split(";")
                if (name := router.strip())
            ]
        # Create FastAPI instance
        app_instance = FastAPI(**kwargs)
        return cls(
            app=app_instance,
            include_routers=include_routers,
            exclude_routers=exclude_routers,
        )

    def __post_init__(self):
        logger.info("Attempting to register routers...")
        logger.info("Included routers candidates: %s", self.include_routers or "ALL")
        logger.info("Excluded routers candidates: %s", self.exclude_routers or "NONE")
        for router in RouterCatalog:
            name = router.name.upper()
            if self.include_routers and name not in self.include_routers:
                logger.info(f"Skipping router '{name}' as it's not in the include list.")
                continue
            if self.exclude_routers and name in self.exclude_routers:
                logger.info(f"Skipping router '{name}' as it's in the exclude list.")
                continue
            logger.info(f"Registering router '{name}'.")
            self.app.include_router(
                router.get_router_instance(),
                **router.get_router_configs(),
            ) 

    def start(self, **kwargs):
        import uvicorn

        server_kwargs = {
            "host": "localhost",
            "port": 8000,
            "log_level": "info",
            **kwargs,
        }

        uvicorn.run(self.app, **server_kwargs)



from dataclasses import dataclass, field
from functools import wraps
from typing import Any, Callable, Optional

from fred.settings import logger_manager

logger = logger_manager.get_logger(name=__name__)


# Note: We need to turn off frozen & slots here to allow dynamic attribute setting
# by the wraps decorator
@dataclass(frozen=False, slots=False)
class DBUtilsRetrievalMethod:
    position: int
    function: Callable

    @classmethod
    def with_position(cls, value: int) -> Callable:
        def decorator(function: Callable) -> 'DBUtilsRetrievalMethod':
            instance = cls(position=value, function=function)
            wraps(function)(instance)
            return instance
        return decorator

    def __call__(self, *args, **kwargs) -> Callable:
        return self.function(*args, **kwargs)


@dataclass(frozen=False, slots=True)
class DBUtilsFinder:
    instance: Optional[Any] = None

    
    @staticmethod
    @DBUtilsRetrievalMethod.with_position(value=0)
    def get_dbutils_from_python_context(**kwargs):
        try:
            return (
                locals().get("dbutils")
                or globals().get("dbutils")
            )
        except Exception as e:
            logger.warning(f"Failed to get dbutils from python context: {e}")
            return None

    @staticmethod
    @DBUtilsRetrievalMethod.with_position(value=1)
    def get_dbutils_from_sdk_context(**kwargs):
        try:
            from databricks.sdk.runtime import dbutils
            return dbutils
        except ImportError:
            from databricks.sdk.runtime import get_dbutils
            return get_dbutils()
        except Exception as e:
            logger.warning(f"Failed to get dbutils from SDK context: {e}")
            return None

    @staticmethod
    @DBUtilsRetrievalMethod.with_position(value=2)
    def get_dbutils_from_workspace_client(client: Optional['WorkspaceClient'] = None, **kwargs):
        try:
            if client is None:
                from databricks.sdk import WorkspaceClient
                client = WorkspaceClient()
            return client.dbutils
        except Exception as e:
            logger.warning(f"Failed to get dbutils from workspace client: {e}")
            return None
        
    @staticmethod
    @DBUtilsRetrievalMethod.with_position(value=3)
    def get_dbutils_from_pypspark_import(spark: Optional['SparkSession'] = None, **kwargs):
        try:
            from pyspark.dbutils import DBUtils  # type: ignore[import]
            return DBUtils(spark)
        except Exception as e:
            logger.warning(f"Failed to get dbutils from pyspark context: {e}")
            return None
        
    @staticmethod
    @DBUtilsRetrievalMethod.with_position(value=4)
    def get_dbutils_from_jvm_context(spark: Optional['SparkSession'] = None, **kwargs):
        try:
            if spark is None:
                from pyspark.sql import SparkSession
                spark = SparkSession.builder.getOrCreate()
            return spark._jvm.dbutils
        except Exception as e:
            logger.warning(f"Failed to get dbutils from JVM context: {e}")
            return None
        
    @classmethod
    def find_dbutils(
        cls,
        spark: Optional['SparkSession'] = None,
        client: Optional['WorkspaceClient'] = None,
        start_with_index: int = 0,
    ):
        get_dbutils_methods = sorted(
            [
                method
                for method in dir(cls)
                if method.startswith("get_dbutils_from_")
            ],
            key=lambda m: getattr(cls, m).position
        )

        def try_methods(methods: list[str], index: int = 0):
            # Early exit if index is out of bounds
            if index >= len(methods):
                return None
            # Recursively try each method until one succeeds or we run out of options
            get_dbutils_method, *others = methods[index:]
            if (dbutils := getattr(cls, get_dbutils_method)(spark=spark, client=client)):
                return dbutils
            return try_methods(others, index + 1) if others else None

        return try_methods(methods=get_dbutils_methods, index=start_with_index)
    
    def get(self, **kwargs):
        if self.instance:
            return self.instance
        self.instance = self.find_dbutils(**kwargs)
        return self.instance
    

@dataclass(frozen=True, slots=True)
class DBUtilsWrapper:
    instance: Optional[Any] = None

    @classmethod
    def auto(
        cls,
        spark: Optional['SparkSession'] = None,
        client: Optional['WorkspaceClient'] = None,
        start_with_index: int = 0,
    ) -> 'DBUtilsWrapper':
        return cls(
            instance=DBUtilsFinder.find_dbutils(
                spark=spark,
                client=client,
                start_with_index=start_with_index,
            )
        )



from threading import Thread
from typing import (
    Callable,
    Optional,
    TypeVar,
)

from fred.settings import logger_manager
from fred.future.settings import FRD_FUTURE_DEFAULT_EXPIRATION
from fred.future.callback.interface import CallbackInterface
from fred.monad.interface import MonadInterface
from fred.monad.catalog import EitherMonad
from fred.future.result import (
    FutureResult,
    FutureUndefinedPending,
    FutureUndefinedInProgress,
    FutureDefined,
)

A = TypeVar("A")
B = TypeVar("B")

logger = logger_manager.get_logger(__name__)


class Future(MonadInterface[A]):
    """A Future represents a computation that will complete at some point in the future,
    yielding a result of type A or failing with an exception. It allows for asynchronous
    programming by enabling non-blocking operations and chaining of computations.
    This implementation uses threading to execute the computation in a separate thread.

    The Future class provides methods to wait for the computation to complete,
    retrieve the result, and chain further computations using flat_map and map.

    Note: This implementation is a simplified version and may not cover all edge cases
    or provide advanced features found in more comprehensive concurrency libraries.

    Note: The Future class is designed to work with the Either monad to encapsulate
    successful results (Right) and exceptions (Left).

    Note: The Future class uses a backend storage system to persist the state
    and result of the computation, allowing for distributed or long-running tasks.

    Note: The Future class is generic and can be used with any type A, allowing for
    flexibility in the types of computations it can represent.

    Note: Even though the Future class implements the MonadInterface, it is not a full monad
    in the traditional sense, as it represents an asynchronous computation rather than
    a pure value. However, it provides similar capabilities for chaining and transforming
    computations in a functional style.

    TODO: Consider adding cancellation support to allow aborting ongoing computations.
    
    TODO: Analyze impact on performance and resource usage when using many Futures. Specially
    on garbage collection and memory leaks.
    
    TODO: Consider using more advanced concurrency primitives for better control and efficiency.
    """

    def __init__(
            self,
            function: Callable[..., A],
            on_start: Optional[CallbackInterface] = None,
            on_complete: Optional[CallbackInterface] = None,
            **kwargs
        ):
        """Initializes a Future with the provided function to be executed asynchronously.
        The function is executed in a separate thread, allowing for non-blocking operations.

        Args:
            function (Callable[..., A]): The function to be executed asynchronously.
            future_id (Optional[str], keyword-only): A reserved keyword-only parameter used to
                uniquely identify this Future instance. This value is consumed by the Future
                infrastructure and is not passed to the target function.
            **kwargs: Additional keyword arguments to be passed to the function when executed.
                All keyword arguments except 'future_id' are forwarded to the target function.
        """
        # Create a new available future
        future = FutureUndefinedPending.auto(future_id=kwargs.pop("future_id", None))
        # Register the Future-ID and define the available future via the provided function.
        # Note: The 'apply' method is blocking by itself; thus, we run it in a separate thread.
        # Note: The thread is a daemon to ensure it does not block program exit.
        self.future_id = future.future_id
        self.thread = Thread(
            target=lambda: future.apply(
                function=function,
                on_complete=on_complete,
                on_start=on_start,
                **kwargs
            ),
            daemon=True,
        )
        # Start the thread to execute the function asynchronously
        self.thread.start()

    @property
    def _result(self) -> Optional[FutureResult[A]]:
        return FutureResult.from_backend(future_id=self.future_id)
    
    @property
    def _status(self) -> Optional[str]:
        return FutureResult._get_status_key(future_id=self.future_id).get()

    @property
    def _output(self) -> Optional[str]:
        return FutureResult._get_output_key(future_id=self.future_id).get()

    def wait(self, timeout: Optional[float] = None) -> EitherMonad.Either[A]:
        """Waits for the future to complete and returns the result as an Either monad.

        TL;DR Waiting will collapse the future into a concrete result
        which can be either a value (Right) or an exception (Left).

        * If the future completes successfully, returns Right(value).
        * If the future fails, returns Left(exception).
        * If the timeout is reached before completion, raises a TimeoutError.
        Args:
            timeout (Optional[float]): Maximum time to wait for the future to complete.
                                       If None, waits indefinitely.
        Returns:
            Either[A]: An Either monad containing the result or exception.
        Raises:
            TimeoutError: If the future does not complete within the specified timeout.
            RuntimeError: If the future status is inconsistent after waiting.
            TypeError: If the future result type is unknown.
        """
        # Wait for the thread to complete or timeout
        self.thread.join(timeout=timeout)
        if self.thread.is_alive():
            raise TimeoutError("Future did not complete within the specified timeout.")
        # After the thread has completed, check the status consistency
        if not self._status:
            raise RuntimeError("Future status should be set but isn't.")
        if not self._status.startswith("DEFINED"):
            raise RuntimeError("Future status should be DEFINED after wait.")
        # Check the result-type consistency
        # If the future is still pending or in-progress, it's an error and should not happen
        # If the future is defined, return the contained value or exception
        match self._result:
            case FutureUndefinedPending():
                raise RuntimeError("Future is still pending after wait.")
            case FutureUndefinedInProgress():
                raise RuntimeError("Future is still in progress after wait.")
            case FutureDefined(value=value):
                return value
            case _:
                raise TypeError("Unknown FutureResult type")

    def getwhatevernow(self) -> Optional[EitherMonad.Either[A]]:
        """Gets the current result of the future without waiting.
        If the future is not yet completed, returns None.
        If the future is completed, returns the result as an Either monad.

        Returns:
            Optional[Either[A]]: None if the future is not completed; otherwise, an Either monad containing the result or exception.
        """
        match self._result:
            case FutureDefined(value=value):
                return value
            case _:
                return None

    @property
    def state(self) -> Optional[str]:
        """Returns the current state of the future as a string.
        The state is derived from the status key in the backend storage.
        Returns:
            Optional[str]: The current state of the future, or None if not available.
                           Possible states include:
                           - "UNDEFINED:PENDING"
                           - "UNDEFINED:IN_PROGRESS"
                           - "DEFINED:SUCCESS"
                           - "DEFINED:FAILURE"
        """
        return ":".join(self._status.split(":")[:2]) if self._status else None

    def __repr__(self) -> str:
        return f"FUTURE[{self.state}]('{self.future_id}')"
    
    def __str__(self) -> str:
        return self.future_id

    def wait_and_resolve(self, timeout: Optional[float] = None) -> A:
        """Waits for the future to complete and resolves the result into
        the expected output type of the function call; raises if the future failed.
        This is a convenience method that combines waiting for the future to complete
        and resolving the result, raising any exceptions that occurred during execution.
        Args:
            timeout (Optional[float]): Maximum time to wait for the future to complete.
                                       If None, waits indefinitely.
        Returns:
            A: The resolved value of the future if it completed successfully.
        Raises:
            TimeoutError: If the future does not complete within the specified timeout.
            Exception: If the future failed during execution, the original exception is raised.
        """
        return self.wait(timeout=timeout).resolve()

    @classmethod
    def from_value(cls, val: A) -> 'Future[A]':
        return Future(function=lambda: val)
    
    def flat_map(self, function: Callable[[A], 'Future[B]'], timeout: Optional[float] = None) -> 'Future[B]':
        """Chains the current future with another future-producing function.
        This method allows for sequential asynchronous operations where the output
        of one future is used as the input to another.

        The simple implementation of this method is: function(self.wait_and_resolve())
        Nonetheless, the simple implementation would block the current thread
        until the first future is resolved, which is not ideal in an asynchronous context.
        Instead, we create a new Future that encapsulates the entire operation,
        allowing it to be executed asynchronously.

        Args:
            function (Callable[[A], Future[B]]): A function that takes the result of the
                                                  current future and returns a new Future.
            timeout (Optional[float]): Maximum time to wait for the current future to complete.
                                       If None, waits indefinitely.
        Returns:
            Future[B]: A new Future representing the chained operation."""
        # TODO: Is there a more efficient implementation?
        return Future(
            function=lambda:
                function(self.wait_and_resolve(timeout=timeout))
                .wait_and_resolve(timeout=timeout)
        )

    def map(self, function: Callable[[A], B]) -> 'Future[B]':
        """Applies a function to the result of the future, returning a new future.
        This method allows for transforming the result of a future without blocking.
        Args:
            function (Callable[[A], B]): A function that takes the result of the current future
                                          and returns a new value.
        Returns: Future[B]: A new Future containing the transformed result.
        """
        return self.flat_map(function=lambda value: type(self).from_value(function(value)))

    @classmethod
    def pullsync(
            cls,
            future_id: str,
            retry_delay: float = 0.2,
            retry_backoff_rate: float = 0.1,
            retry_delay_max: float = 15,
            timeout: float = FRD_FUTURE_DEFAULT_EXPIRATION,
            on_complete: Optional[CallbackInterface] = None,
            **kwargs
        ) -> 'Future[A]':
        """Pulls an existing future from the backend storage by its ID.
        This method allows for retrieving and interacting with a future
        that was previously created and stored in the backend.

        Args:
            future_id (str): The unique identifier of the future to be pulled.
            retry_delay (float): Initial delay between checks for the future's completion.
            retry_backoff_rate (float): Incremental increase in delay after each check.
            retry_delay_max (float): Maximum delay between checks.
            timeout (float): Maximum time to wait for the future to complete.
            on_complete (Optional[CallbackInterface]): An optional callback to be executed
                                                      when the future completes.
            **kwargs: Additional keyword arguments to be passed to the Future constructor.
        Returns:
            Future[A]: A Future instance representing the pulled future.
        """
        from fred.future.utils import pull_future_result

        return cls(
            function=lambda: pull_future_result(
                future_id=future_id,
                retry_delay=retry_delay,
                retry_backoff_rate=retry_backoff_rate,
                retry_delay_max=retry_delay_max,
                timeout=timeout,
            ),
            on_complete=on_complete,
            **kwargs
        )



import enum

from fred.worker.runner.model._runner_spec import RunnerSpec
from fred.worker.runner.model._request import RunnerRequest
from fred.worker.runner.model._item import RunnerItem
from fred.worker.runner.model._handler import Handler


class RunnerModelCatalog(enum.Enum):
    RUNNER_SPEC = RunnerSpec
    REQUEST = RunnerRequest
    HANDLER = Handler
    ITEM = RunnerItem

    def __call__(self, *args, **kwargs):
        return self.value(*args, **kwargs)



import enum

from fred.monad._either import EitherMonad


class MonadCatalog(enum.Enum):
    """Enum representing the different types of Monads."""
    EITHER = EitherMonad

    def __call__(self, *args, **kwargs):
        return self.value.from_value(*args, **kwargs)



from dataclasses import dataclass

from fred.settings import logger_manager
from fred.dao.comp.catalog import FredKeyVal, FredQueue
from fred.dao.service.catalog import ServiceCatalog

logger = logger_manager.get_logger(name=__name__)


@dataclass(frozen=True, slots=False)
class RunnerBackend:
    keyval: FredKeyVal
    queue: FredQueue

    @classmethod
    def auto(cls, service_name: str, **kwargs) -> 'RunnerBackend':
        service = ServiceCatalog[service_name.upper()]
        match service:
            case ServiceCatalog.REDIS:
                from fred.dao.service.utils import get_redis_configs_from_payload
                service_kwargs = get_redis_configs_from_payload(kwargs)
            case ServiceCatalog.STDLIB:
                service_kwargs = {}
            case _:
                logger.error(f"Unknown service '{service_name}'... will attempt to use provided kwargs as-is.")
                service_kwargs = kwargs
        components = service.component_catalog(**service_kwargs)
        return cls(
            keyval=components.KEYVAL.value,
            queue=components.QUEUE.value,
        )



from fred.settings import get_environ_variable


def get_redis_configs_from_payload(
        payload: dict,
        keep: bool = False,
    ) -> dict:
    """Extract Redis configuration from the given payload dictionary.
    This function looks for common Redis configuration keys in the payload
    dictionary. If a key is not found, it falls back to environment variables.
    Args:
        payload (dict): The dictionary from which to extract Redis configuration.
        keep (bool): If True, the original keys are retained in the payload. If False, they are removed.
    Returns:
        dict: A dictionary containing Redis configuration parameters.
    """
    host = port = password = db = None
    for host_key in ["host", "redis_host"]:
        if (host := payload.get(host_key) if keep else payload.pop(host_key, None)):
            break
    for port_key in ["port", "redis_port"]:
        if (port := payload.get(port_key) if keep else payload.pop(port_key, None)):
            break
    for password_key in ["password", "redis_password"]:
        if (password := payload.get(password_key) if keep else payload.pop(password_key, None)):
            break
    for db_key in ["db", "redis_db"]:
        if (db := payload.get(db_key) if keep else payload.pop(db_key, None)):
            break
    return {
        "host": host or get_environ_variable(name="REDIS_HOST", default="localhost"),
        "port": int(port or get_environ_variable(name="REDIS_PORT", default=6379)),
        "password": password or get_environ_variable(name="REDIS_PASSWORD", default=None),
        "db": int(db or get_environ_variable(name="REDIS_DB", default=0)),
        "decode_responses": True,
        **(payload.get("redis_configs", {}) if keep else payload.pop("redis_configs", {})),
    }



from typing import Optional

from fred.settings import get_environ_variable, logger_manager

logger = logger_manager.get_logger(name=__name__)


def get_queue_name_from_payload(
        payload: dict,
        search_keys: list[str],
        env_fallback: str,
        default: Optional[str] = None,
        keep: bool = False,
    ) -> str:
    for key in search_keys:
        if (queue_name := payload.get(key) if keep else payload.pop(key, None)):
            return queue_name
    return (
        queue_name
        or get_environ_variable(name=env_fallback, default=None)
        or (
            logger.warning(f"Redis queue not specified; defaulting to: '{default}'")
            or default
        )
    )

def get_request_queue_name_from_payload(
        payload: dict,
        keep: bool = False,
) -> Optional[str]:
    return get_queue_name_from_payload(
        payload=payload,
        search_keys=["redis_request_queue", "request_queue", "req_queue"],
        env_fallback="FRD_RUNNER_REQUEST_QUEUE",
        default="req:demo",
        keep=keep,
    )

def get_response_queue_name_from_payload(
        payload: dict,
        keep: bool = False,
) -> Optional[str]:
    return get_queue_name_from_payload(
        payload=payload,
        search_keys=["redis_response_queue", "response_queue", "res_queue"],
        env_fallback="FRD_RUNNER_RESPONSE_QUEUE",
        default=None,
        keep=keep,
    )

def get_redis_configs_from_payload(
        payload: dict,
        keep: bool = False,
    ) -> dict:
    for host_key in ["host", "redis_host"]:
        if (host := payload.get(host_key) if keep else payload.pop(host_key, None)):
            break
    for port_key in ["port", "redis_port"]:
        if (port := payload.get(port_key) if keep else payload.pop(port_key, None)):
            break
    for password_key in ["password", "redis_password"]:
        if (password := payload.get(password_key) if keep else payload.pop(password_key, None)):
            break
    for db_key in ["db", "redis_db"]:
        if (db := payload.get(db_key) if keep else payload.pop(db_key, None)):
            break
    return {
        "host": host or get_environ_variable(name="REDIS_HOST", default="localhost"),
        "port": int(port or get_environ_variable(name="REDIS_PORT", default=6379)),
        "password": password or get_environ_variable(name="REDIS_PASSWORD", default=None),
        "db": int(db or get_environ_variable(name="REDIS_DB", default=0)),
        "decode_responses": True,
        **(payload.get("redis_configs", {}) if keep else payload.pop("redis_configs", {})),
    }



from typing import (
    Callable,
    Generic,
    TypeVar,
)

A = TypeVar("A")
B = TypeVar("B")


class MonadInterface(Generic[A]):
    
    @classmethod
    def from_value(cls, val: A) -> 'MonadInterface[A]':
        raise NotImplementedError
    
    def flat_map(self, function: Callable[[A], 'MonadInterface[B]']) -> 'MonadInterface[B]':
        raise NotImplementedError

    def map(self, function: Callable[[A], B]) -> 'MonadInterface[B]':
        # Map can be implemented using flat_map to avoid code duplication and ensure consistency
        return self.flat_map(function=lambda value: type(self).from_value(function(value)))



from typing import (
    Any,
    Callable,
    TypeVar,
)

from fred.future.callback.interface import CallbackInterface
from fred.monad.catalog import EitherMonad

A = TypeVar("A")


class CallbackFunction(CallbackInterface[A]):
    
    def __init__(self, function: Callable[[EitherMonad.Either[A]], None], **kwargs):
        self.function = function
        self.kwargs = kwargs

    def _on_start(self, future_id: str) -> Any:
        return self.function(
            future_id=future_id,
            **self.kwargs
        )

    def _on_complete(self, future_id: str, output: EitherMonad.Either[A]) -> Any:
        return self.function(
            output=output,
            future_id=future_id,
            **self.kwargs
        )



import uuid
import json
from functools import lru_cache
from typing import Generic, TypeVar

T = TypeVar("T")


class ServiceConnectionPoolInterface(Generic[T]):
    pool_registry: dict[str, T] = {}

    @classmethod
    def get_pool_id(cls, **kwargs) -> str:
        return str(uuid.uuid5(uuid.NAMESPACE_OID, json.dumps(kwargs, sort_keys=True)))

    @classmethod
    def _create_pool(cls, **kwargs) -> T:
        raise NotImplementedError("This method should be implemented by subclasses.")

    @classmethod
    @lru_cache
    def get_or_create_pool(cls, **kwargs) -> T:
        pool_id = cls.get_pool_id(**kwargs)
        pool = cls.pool_registry.get(pool_id)
        if not pool:
            pool = cls.pool_registry[pool_id] = cls._create_pool(**kwargs)
        return pool


class ServiceInterface(Generic[T]):
    instance: T

    def __init__(self, **kwargs):
        self.config = kwargs

    @classmethod
    def _create_instance(cls, **kwargs) -> T:
        raise NotImplementedError("This method should be implemented by subclasses.")
    
    @classmethod
    def auto(cls, **kwargs) -> "ServiceInterface":
        raise NotImplementedError("This method should be implemented by subclasses.")

    @property
    def client(self) -> T:  
        if not getattr(self, "instance", None):
            self.instance = self._create_instance(**getattr(self, "config", {}))
        return self.instance
    
    def close(self):
        # Close the instance if it has a close method.
        # This method can be overridden by subclasses if needed.
        if self.instance and hasattr(self.instance, "close") and callable(self.instance.close):
            self.instance.close()



from queue import Queue
from dataclasses import dataclass
from typing import Optional

from fred.utils.runtime import RuntimeInfo
from fred.dao.service.interface import ServiceConnectionPoolInterface, ServiceInterface


class StdLibSingleton:
    _instance: Optional["StdLib"] = None

    @classmethod
    def get_instance(cls, **kwargs) -> "StdLib":
        if cls._instance is None:
            cls._instance = StdLib.auto(**kwargs)
        return cls._instance


@dataclass(frozen=True, slots=True)
class StdLib:
    runtime_info: RuntimeInfo
    _memstore_keyval: dict[str, str]
    _memstore_queue: dict[str, Queue]

    @classmethod
    def auto(cls, **kwargs) -> "StdLib":
        _memstore_keyval = kwargs.pop("memstore_keyval", {})
        _memstore_queue = kwargs.pop("memstore_queue", {})
        return cls(
            runtime_info=RuntimeInfo.auto(**kwargs),
            _memstore_keyval=_memstore_keyval,
            _memstore_queue=_memstore_queue,
        )


class StdLibConnectionPool(ServiceConnectionPoolInterface[StdLibSingleton]):

    @classmethod
    def _create_pool(cls, **kwargs) -> StdLibSingleton:
        return StdLibSingleton(**kwargs)


class StdLibService(ServiceInterface[StdLib]):
    instance: StdLib

    @classmethod
    def _create_instance(cls, **kwargs) -> StdLib:
        pool = StdLibConnectionPool.get_or_create_pool(**kwargs)
        return pool.get_instance(**kwargs)

    @classmethod
    def auto(cls, **kwargs) -> "StdLib":
        cls.instance = cls._create_instance(**kwargs)
        return cls(**kwargs)



from dataclasses import dataclass
from fred.worker.runner.rest.routers.interface import RouterInterface


@dataclass(frozen=True, slots=False)
class RunnerRouter(RouterInterface):

    def __post_init__(self):
        self.router.add_api_route(
            "/handler_exists",
            self.handler_exists,
            methods=["GET"],
            tags=["Runner"],
            summary="Check if a handler class exists and is a RunnerHandler.",
            response_description="Details about the handler class.",
        )

    def handler_exists(self, classname: str, classpath: str) -> dict:
        from fred.worker.runner.handler import RunnerHandler
        from fred.worker.interface import HandlerInterface

        result_payload = {
            "handler_classname": classname,
            "handler_classpath": classpath,
            "exists": False,
            "is_runner_handler": False,
            "metadata": {}
        }

        try:
            handler = HandlerInterface.find_handler(
                import_pattern=classpath,
                handler_classname=classname,
            )
            result_payload["is_runner_handler"] = isinstance(handler, RunnerHandler)
            result_payload["exists"] = True
            return result_payload
        except Exception as e:
            result_payload["metadata"]["error"] = str(e)
            return result_payload



from fred.worker.interface import HandlerInterface
from fred.cli.interface import IntegrationExtCLI
from fred.settings import logger_manager


logger = logger_manager.get_logger(name=__name__)


class RunPodExt(IntegrationExtCLI):

    def get_handler_instance(self, import_pattern: str, handler_classname: str) -> HandlerInterface:
        return HandlerInterface.find_handler(
            import_pattern=import_pattern,
            handler_classname=handler_classname,
        )

    def execute_local(self, import_pattern: str, handler_classname: str, **kwargs) -> dict:
        payload = kwargs.pop("payload", {})
        handler = self.get_handler_instance(
            import_pattern=import_pattern,
            handler_classname=handler_classname,
        )
        return handler.run(
            event={
                "id": "local-exec",
                "input": payload,
            }
        )

    def execute(self, import_pattern: str, handler_classname: str, local: bool = False, **kwargs):
        # Early exit and redirect to local execution when specified.
        if local:
            return self.execute_local(import_pattern, handler_classname, **kwargs)
        
        # Lazy import to avoid dependency issues when not using RunPod
        import runpod  # type: ignore
        
        logger.info(f"Starting RunPod serverless with handler '{handler_classname}' from '{import_pattern}'.")
        handler = self.get_handler_instance(
            import_pattern=import_pattern,
            handler_classname=handler_classname,
        )
        runpod.serverless.start(
            {
                "handler": handler.run,
                **kwargs,
            }
        )



from typing import Optional
from fred.dao.service.interface import ServiceInterface
from fred.dao.service.catalog import ServiceCatalog

SRV_REF_TYPE = str | ServiceInterface | ServiceCatalog


class SrvCompanionMixin:
    _srv: ServiceInterface

    @classmethod
    def _set_srv(cls, srv_ref: Optional[SRV_REF_TYPE] = None, **kwargs):
        """Sets the service instance for the component class.
        This method initializes the `_srv` class variable with an instance
        of the service specified by `name` and any additional parameters passed via `kwargs`.
        Args:
            name (Optional[str]): The name of the service to set. Defaults to "REDIS".
            **kwargs: Additional keyword arguments to configure the service instance.
        """
        match (srv_ref or "REDIS"):
            case str() as name:
                cls._srv = ServiceCatalog[name.upper()].auto(**kwargs)
            case ServiceCatalog() as cat:
                cls._srv = cat.auto(**kwargs)
            case ServiceInterface() as instance:
                cls._srv = instance
            case _:
                raise ValueError(f"Invalid service '{srv_ref}' type: {type(srv_ref)}")
    
    @property
    def _nme(self) -> str:
        """Returns the class name of the current service instance."""
        return self._srv.__class__.__name__

    @property
    def _cat(self) -> ServiceCatalog:
        """Returns the ServiceCatalog enum member corresponding to the current service instance."""
        return ServiceCatalog.from_classname(self._nme)


class ComponentInterface(SrvCompanionMixin):

    @classmethod
    def mount(cls, srv_ref: Optional[SRV_REF_TYPE] = None, **kwargs) -> type["ComponentInterface"]:
        """Mounts the component to a specific service instance.
        This method configures the component to use a service instance
        identified by `srv_name` and any additional parameters passed via `kwargs`.
        
        Args:
            srv_name (Optional[str]): The name of the service to mount. Defaults to "REDIS".
            **kwargs: Additional keyword arguments to configure the service instance.
        """
        cls._set_srv(srv_ref=srv_ref, **kwargs)
        return cls

    @classmethod
    def auto(cls, srv_ref: Optional[SRV_REF_TYPE] = None, **kwargs) -> "ComponentInterface":
        """Automatically creates an instance of the component, mounting it to a service.
        This method is a convenience wrapper that first mounts the component to a service
        and then creates an instance of the component.
        Args:
            srv_name (Optional[str]): The name of the service to mount. Defaults to "REDIS".
            **kwargs: Additional keyword arguments for both mounting the service and
                      creating the component instance. If there are specific arguments
                      for the service, they should be passed under the key `srv_kwargs`
                      as a dictionary.
        """
        srv_kwargs = kwargs.pop("srv_kwargs", {})
        return cls.mount(srv_ref=srv_ref, **srv_kwargs)(**kwargs)



from dataclasses import dataclass

from fred.settings import logger_manager
from fred.worker.interface import HandlerInterface
from fred.worker.runner.model.interface import ModelInterface

logger = logger_manager.get_logger(name=__name__)


@dataclass(frozen=True, slots=False)
class Handler(ModelInterface):
    classname: str
    classpath: str
    kwargs: dict

    def get_handler(self) -> HandlerInterface:
        return HandlerInterface.find_handler(
            handler_classname=self.classname,
            import_pattern=self.classpath,
            **self.kwargs,
        )



from threading import Thread
from typing import (
    Generic,
    TypeVar,
    Optional,
)

from fred.settings import logger_manager
from fred.monad.catalog import EitherMonad

logger = logger_manager.get_logger(__name__)

A = TypeVar("A")


class CallbackInterface(Generic[A]):
    
    def _on_start(self, future_id: str):
        raise NotImplementedError
    
    def _on_complete(self, future_id: str, output: EitherMonad.Either[A]):
        raise NotImplementedError
    
    def run(self, future_id: str, blocking: bool = False, output: Optional[EitherMonad.Either[A]] = None) -> Optional[Thread]:
        """Executes the callback with the provided output and handles any exceptions.
        Args:
            output (EitherMonad.Either[A]): The output to be passed to the callback.
        Returns:
            bool: True if the callback executed successfully, False otherwise.
        """
        # TODO: Consider using a richer return type to capture more details about the execution
        #  and optionally propagate the callback return value.
        try:
            match output:
                case None:
                    thread = Thread(
                        target=lambda: self._on_start(future_id=future_id),
                        daemon=True
                    )
                case EitherMonad.Either():
                    thread = Thread(
                        target=lambda: self._on_complete(future_id=future_id, output=output),
                        daemon=True
                    )
            thread.start()
            if blocking:
                thread.join()
            return thread
        except Exception as e:
            logger.error(f"Callback execution failed on future '{future_id}': {e}")
            return None



import enum
from dataclasses import dataclass

from fred.settings import logger_manager
from fred.worker.runner.plugins._local import LocalPlugin

logger = logger_manager.get_logger(name=__name__)


class PluginCatalog(enum.Enum):
    """Enum for the different plugins available in FRED."""

    LOCAL = LocalPlugin
    RUNPOD = None  # Placeholder for future RunPod plugin
    LAMBDA = None  # Placeholder for future AWS Lambda plugin

    def __call__(self, *args, **kwargs):
        return self.value.auto(*args, **kwargs)



from dataclasses import dataclass
from typing import Optional

from fastapi import APIRouter


@dataclass(frozen=True, slots=False)
class RouterInterface:
    router: APIRouter
    router_configs: dict

    @classmethod
    def auto(
        cls,
        router: Optional[APIRouter] = None,
        **kwargs,
    ) -> "RouterInterface":
        return cls(
            router=router or APIRouter(),
            router_configs=kwargs,
        )

    def __post_init__(self):
        self.router.add_api_route(
            "/ping",
            self.ping,
            methods=["GET"],
            tags=["Health"],
            summary="Ping the server to check if it's alive.",
            response_description="A simple pong response.",
        )

    def ping(self, pong: Optional[str] = None) -> dict:
        from fred.utils.dateops import datetime_utcnow

        return {
            "ping_time": datetime_utcnow().isoformat(),
            "ping_response": pong or "pong",
        }



import os
import enum
import logging
from typing import Callable, Dict, Literal, NoReturn, Optional, Union, TypeVar, overload
from logging.config import dictConfig

T = TypeVar("T")


@overload
def get_environ_variable(
        name: str,
        default: Literal[None] = None,
        enforce: bool = False,
        apply: Optional[Callable[[Optional[str]], T]] = None,
) -> Optional[Union[str, T]]:
    ...


@overload
def get_environ_variable(
        name: str,
        default: str,
        enforce: bool = False,
        apply: Optional[Callable[[str], T]] = None,
) -> Union[str, T]:
    ...


def get_environ_variable(
        name: str,
        default: Optional[str] = None,
        enforce: bool = False,
        apply: Optional[Union[Callable[[Optional[str]], T], Callable[[str], T]]] = None,
) -> Optional[Union[Optional[str], T]]:
    return (apply or (lambda x: x))(  # type: ignore
        os.environ.get(name, default=default) if not enforce else  # type: ignore
        os.environ.get(name) or (lambda: (_ for _ in ())
                                 .throw(ValueError(f"Missing environ variable: {name}")))()
    )


# Environment variables for the OpenAI API
FRD_OPENAI_API_KEY = get_environ_variable(
    name="FRD_OPENAI_API_KEY",
    enforce=False,
)
FRD_OPENAI_BASE_URL = get_environ_variable(
    name="FRD_OPENAI_BASE_URL",
    default="https://api.openai.com/v1"
)
FRD_OPENAI_DEFAULT_MODEL = get_environ_variable(
    "FRD_OPENAI_DEFAULT_MODEL",
    default="openai/gpt-oss-20b"
)


# Logger configuration
default_logger_configuration = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {
            "format": get_environ_variable(
                name="DEFAULT_PYTHON_LOGGER",
                default="%(asctime)s [%(levelname)s] %(name)s: %(message)s"
            )
        },
    },
    "handlers": {
        "default": {
            "level": get_environ_variable(
                name="DEFAULT_PYTHON_HANDLER_LEVEL",
                default="INFO"
            ),
            "formatter": "standard",
            "class": "logging.StreamHandler",
        },
    },
    "loggers": {
        "": {
            "handlers": [
                "default"
            ],
            "level": get_environ_variable(
                name="DEFAULT_PYTHON_LOGGER_LEVEL",
                default="INFO",
            ),
            "propagate": True
        }
    }
}


def configure_logger_callable(config_dictionary: Optional[Dict] = None) -> Callable:
    def get_logger(name: str):
        dictConfig(config_dictionary or default_logger_configuration)
        return logging.getLogger(name)
    return get_logger


class LoggerManager:
    class _Singleton:
        count: int = 0
        value: Optional['LoggerManager'] = None

    def __new__(cls, *args, **kwargs):
        if cls._Singleton.count == 0:
            instance = super(LoggerManager, cls).__new__(cls)
            cls._Singleton.count += 1
            cls._Singleton.value = instance
            return instance
        return cls._Singleton.value

    def __init__(self, config_dictionary: Optional[Dict] = None, disable_singleton: bool = False):
        self.config_dictionary: Dict = config_dictionary or default_logger_configuration

        if not disable_singleton and self._Singleton.count > 0:
            return
        self._Singleton.count += 1
        self._Singleton.value = self

    def set_configuration(self, **kwargs):
        self.config_dictionary = kwargs

    def get_logger(self, name: str, overwrite_config_dictionary: Optional[Dict] = None):
        dictConfig(overwrite_config_dictionary or self.config_dictionary)
        return logging.getLogger(name)

    @classmethod
    def singleton(cls, overwrite_config_dictionary: Optional[Dict] = None) -> 'LoggerManager':
        if overwrite_config_dictionary or cls._Singleton.value is None:
            cls._Singleton.value = cls(overwrite_config_dictionary)
            if cls._Singleton.count > 1:
                cls._Singleton.value.get_logger(name=__name__).warning("Creating a new logger manager instance.")
        return cls._Singleton.value


logger_manager = LoggerManager.singleton()

if "openrouter" in FRD_OPENAI_BASE_URL and not FRD_OPENAI_DEFAULT_MODEL.endswith(":free"):
    logger_manager.get_logger(__name__).warning("Using OpenRouter with a non-free model.")



from enum import Enum, auto

from fred.utils.dateops import datetime_utcnow


class RunnerStatus(Enum):
    """Statuses used for tracking the runner process."""
    UNDEFINED = auto()
    STARTED = auto()
    STOPPED = auto()
    RUNNING = auto()
    ERROR = auto()

    @staticmethod
    def get_key(runner_id: str) -> str:
        """Get the status key for the specified runner ID."""
        return f"runner:status:{runner_id}"
    
    def get_val(self, *include) -> str:
        """Get the string representation of the status."""
        return ":".join([self.name, datetime_utcnow().isoformat(), *include]) 
    
    @classmethod
    def parse_value(cls, value: str) -> "RunnerStatus":
        """Parse the status from the stored value."""
        val, *_ = value.split(":")
        return cls[val.upper()]



import uuid
import json
from typing import Optional

from pydantic.dataclasses import dataclass

from fred.dao.comp.catalog import FredQueue
from fred.worker.runner.model.interface import ModelInterface


@dataclass(frozen=True, slots=True)
class RunnerRequest(ModelInterface):
    request_id: str
    payload: dict

    @classmethod
    def uuid(
            cls,
            payload: dict,
            uuid_hash: bool = False,
            request_id: Optional[str] = None,
    ) -> "RunnerRequest":
        request_id = request_id or (
            str(uuid.uuid5(uuid.NAMESPACE_OID, json.dumps(payload, default=str)))
            if uuid_hash else str(uuid.uuid4())
        )
        return cls(request_id=request_id, payload=payload)
    
    def as_payload(self) -> dict:
        return {
            "request_id": self.request_id,
            **self.payload,
        }
    
    def dispatch(self, request_queue: FredQueue, **kwargs):
        serialization_kwargs = {
            "default": str,
            **kwargs
        }
        request = json.dumps(self.as_payload(), **serialization_kwargs)
        return request_queue.add(request)



import time
import datetime as dt
from typing import Optional
from dataclasses import dataclass, field

from fred.settings import (
    get_environ_variable,
    logger_manager,
)

logger = logger_manager.get_logger(name=__name__)


@dataclass(slots=True, frozen=True)
class AbstractCLI:
    start_counter: float = field(default_factory=time.perf_counter)
    start_ts: str = field(default_factory=dt.datetime.utcnow().isoformat)

    def now(self, local: bool = False) -> str:
        return (dt.datetime.utcnow() if not local else dt.datetime.now()).isoformat()

    def environ(self, name: str) -> Optional[str]:
        return get_environ_variable(
            name=name,
            enforce=False,
        )

    def runtime(
            self,
            include_modules: bool = False,
        ) -> dict:
        from fred.utils.runtime import RuntimeInfo

        return RuntimeInfo.auto().to_dict(exclude_modules=not include_modules)

    def on_start(self):
        logger.debug("CLI Method not implemented: on_start")

    def on_finalize(self):
        logger.debug("CLI method not implemented: on_finalize")

    def __enter__(self) -> 'AbstractCLI':
        self.on_start()
        return self

    def __exit__(self, *args):
        self.on_finalize()
        end_counter = time.perf_counter()
        # TODO: Add warning logging if an error occurred
        logger.debug(f"Command Timestamp Start: {self.start_ts}")
        logger.debug(f"Command Timestamp Finalize: {self.now(local=False)}")
        logger.info(f"Command Duration: {end_counter - self.start_counter}")

    @classmethod
    def default_config(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    @classmethod
    def cli_exec(cls, *args, **kwargs):
        import fire

        with cls.default_config(*args, **kwargs) as cli:
            fire.Fire(cli)


class IntegrationExtCLI:
    pass


from dataclasses import dataclass
from typing import Optional

from fred.settings import logger_manager
from fred.dao.service.catalog import ServiceCatalog
from fred.dao.comp.interface import ComponentInterface

logger = logger_manager.get_logger(name=__name__)


@dataclass(frozen=True, slots=True)
class FredQueue(ComponentInterface):
    """A simple queue implementation using a backend service.
    This class provides methods to interact with a queue, such as adding,
    removing, and checking the size of the queue. The actual implementation
    of these methods depends on the underlying service being used (e.g., Redis).
    Attributes:
        name: str: The name of the queue.
    """
    name: str

    def size(self) -> int:
        """Returns the number of items in the queue.
        The implementation of this method depends on the underlying service.
        For example, if the service is Redis, it uses the LLEN command to get the
        length of the list representing the queue.
        Returns:
            int: The number of items in the queue.
        Raises:
            NotImplementedError: If the method is not implemented for the current service.
        """
        match self._cat:
            case ServiceCatalog.REDIS:
                return self._srv.client.llen(self.name)
            case ServiceCatalog.STDLIB:
                q = self._srv.client._memstore_queue.get(self.name, None)
                return q.qsize() if q else 0
            case _:
                raise NotImplementedError(f"Size method not implemented for service {self._nme}")

    def clear(self) -> None:
        """Clears all items from the queue.
        The implementation of this method depends on the underlying service.
        For example, if the service is Redis, it uses the DEL command to remove the
        key representing the queue.
        Raises:
            NotImplementedError: If the method is not implemented for the current service.
        """
        match self._cat:
            case ServiceCatalog.REDIS:
                self._srv.client.delete(self.name)
            case ServiceCatalog.STDLIB:
                if (q := self._srv.client._memstore_queue.pop(self.name, None)):
                    del q
            case _:
                raise NotImplementedError(f"Clear method not implemented for service {self._nme}")

    def add(self, item: str) -> None:
        """Adds an item to the queue.
        The implementation of this method depends on the underlying service.
        For example, if the service is Redis, it uses the LPUSH command to add the
        item to the front of the list representing the queue.
        Args:
            item (str): The item to add to the queue.
        Raises:
            NotImplementedError: If the method is not implemented for the current service.
        """
        match self._cat:
            case ServiceCatalog.REDIS:
                self._srv.client.lpush(self.name, item)
            case ServiceCatalog.STDLIB:
                from queue import Queue
                qs = self._srv.client._memstore_queue
                q = qs[self.name] = qs.get(self.name, None) or Queue()
                q.put(item)
            case _:
                raise NotImplementedError(f"Add method not implemented for service {self._srv._nme}")

    def pop(self) -> Optional[str]:
        """Removes and returns an item from the queue.
        The implementation of this method depends on the underlying service.
        For example, if the service is Redis, it uses the RPOP command to remove and
        return the last item from the list representing the queue.
        Returns:
            Optional[str]: The item removed from the queue, or None if the queue is empty.
        Raises:
            NotImplementedError: If the method is not implemented for the current service.
        """
        match self._cat:
            case ServiceCatalog.REDIS:
                return self._srv.client.rpop(self.name)
            case ServiceCatalog.STDLIB:
                from queue import Empty
                if not (q := self._srv.client._memstore_queue.get(self.name, None)):
                    return None
                try:
                    return q.get_nowait()
                except Empty:
                    logger.info(f"Queue '{self.name}' is empty.")
                    return None
            case _:
                raise NotImplementedError(f"Pop method not implemented for service {self._srv._nme}")



from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class DatabricksRuntimeScanner:
    target_url: str

    @classmethod
    def default(cls) -> 'DatabricksRuntimeScanner':
        return cls(
            target_url="https://docs.databricks.com/aws/en/release-notes/runtime/"
        )

    def get_payload(self) -> list[dict]:
        # TODO: Extract the runtime table as a json/dict payload
        raise NotImplementedError



import datetime as dt
from typing import Optional


def datetime_now(tz: Optional[dt.tzinfo] = None) -> dt.datetime:
    return dt.datetime.now(tz=tz)

def datetime_utcnow() -> dt.datetime:
    return datetime_now(tz=dt.timezone.utc)



import uuid
from typing import Optional
from dataclasses import dataclass

from fred.settings import logger_manager

logger = logger_manager.get_logger(name=__name__)


@dataclass(frozen=True, slots=True)
class RunnerInfo:
    runner_id: str
    created_at: str
    runner_inner_handler_classname: str
    runner_inner_handler_classpath: str
    lifetime: int
    timeout: int

    @classmethod
    def create(
        cls,
        runner_inner_handler_classname: str,
        runner_inner_handler_classpath: str,
        runner_id: Optional[str] = None,
        created_at: Optional[str] = None,
        lifetime: int = 3600,  # Default to 1 hour
        timeout: int = 10,  # Default to 10 seconds
    ) -> "RunnerInfo":
        from fred.utils.dateops import datetime_utcnow
        return cls(
            runner_id=runner_id or str(uuid.uuid4()),
            runner_inner_handler_classname=runner_inner_handler_classname,
            runner_inner_handler_classpath=runner_inner_handler_classpath,
            created_at=created_at or datetime_utcnow().isoformat(),
            lifetime=lifetime,
            timeout=timeout,
        )

    def get_start_event(self, **kwargs) -> dict:
        runner_id = kwargs.pop("runner_id", self.runner_id) 
        payload = {
            "lifetime": self.lifetime,
            "timeout": self.timeout,
            "runner_configs": {
                "id": runner_id,
                "import_pattern": self.runner_inner_handler_classpath,
                "handler_classname": self.runner_inner_handler_classname,
            },
            **kwargs,
        }
        return {
            "input": payload,
            "id": runner_id,
        }



import enum
from functools import lru_cache
from typing import Optional

from fred.dao.comp.interface import ComponentInterface, SRV_REF_TYPE
from fred.dao.comp._queue import FredQueue
from fred.dao.comp._keyval import FredKeyVal


class _PreconfCatalogMixin:
    """A mixin class to allow enum members to be called directly to create instances
    of preconfigured components.
    This mixin is used in conjunction with the `preconf` class method of the `CompCatalog`
    enum to create a new enum with preconfigured components for a specific service."""

    def __call__(self, *args, **kwargs) -> ComponentInterface:
        """Create an instance of the preconfigured component. This method
        allows the enum member to be called directly to instantiate the component
        with any additional arguments.
        Args:
            *args: Positional arguments to pass to the component constructor.
            **kwargs: Keyword arguments to pass to the component constructor.
        Returns:
            ComponentInterface: An instance of the preconfigured component.
        """
        return self.value(*args, **kwargs)


class CompCatalog(enum.Enum):
    """An enumeration of available component types.
    Each enum member corresponds to a specific component class that implements
    the `ComponentInterface`. This enum provides a way to reference and create
    instances of different component types in a standardized manner.
    """
    QUEUE = FredQueue
    KEYVAL = FredKeyVal

    @classmethod
    def from_classname(cls, classname: str) -> "CompCatalog":
        """Get enum member by component class name.
        Args:
            classname (str): The class name of the component.
        Returns:
            CompCatalog: The corresponding enum member.
        Raises:
            ValueError: If no matching component is found.
        """
        for item in cls:
            if item.value.__name__ == classname:
                return item
        raise ValueError(f"No component found for classname: {classname}")

    @classmethod
    @lru_cache(maxsize=None)  # TODO: Consider cache invalidation strategy if needed
    def preconf(cls, srv_ref: SRV_REF_TYPE, **kwargs) -> enum.Enum:
        """Create a new Enum with preconfigured components for a specific service name.
        Args:
            srv_name (str): The service name to preconfigure the components with.
            **kwargs: Additional keyword arguments to pass to the component constructors.
        Returns:
            enum.Enum: A new Enum class with preconfigured components.
        """
        from fred.dao.service.interface import ServiceInterface
        from fred.dao.service.catalog import ServiceCatalog
        srv_name = ""
        match srv_ref:
            case str() as name:
                srv_name = name.title()
            case ServiceCatalog() as cat:
                srv_name = cat.name.title()
            case ServiceInterface() as instance:
                srv_name = ServiceCatalog.from_classname(instance.__class__.__name__).name.title()
            case _:
                raise ValueError(f"Invalid service '{srv_ref}' type: {type(srv_ref)}")
        return enum.Enum(
            f"{srv_name.title()}{cls.__name__}",
            {item.name: item.value.mount(srv_ref=srv_ref, **kwargs) for item in cls},
            type=_PreconfCatalogMixin,
        )

    def component_cls(self) -> type[ComponentInterface]:
        """Returns the class of the component associated with the enum member.
        Returns:
            type[ComponentInterface]: The class of the component.
        """
        return self.value

    def auto(self, srv_ref: Optional[SRV_REF_TYPE] = None, **kwargs) -> ComponentInterface:
        """Automatically creates an instance of the component, mounting it to a service.
        This method is a convenience wrapper that first mounts the component to a service
        and then creates an instance of the component.
        Args:
            srv_ref (Optional[SRV_REF_TYPE]): The service reference to mount the component to.
            **kwargs: Additional keyword arguments for both mounting the service and
                      creating the component instance. If there are specific arguments
                      for the service, they should be passed under the key `srv_kwargs`
                      as a dictionary.
        Returns:
            ComponentInterface: An instance of the component.
        """
        srv_ref = srv_ref or kwargs.pop("srv_name", None)  # Backward compatibility
        return self.value.auto(srv_ref=srv_ref, **kwargs)



import time
from typing import overload, Callable, Optional
from dataclasses import dataclass, field

from fred.future.impl import Future  # Should we make this import lazy?
from fred.utils.dateops import datetime_utcnow
from fred.settings import (
    get_environ_variable,
    logger_manager,
)

logger = logger_manager.get_logger(name=__name__)


@dataclass(frozen=True, slots=False)
class HandlerInterface:
    """Base interface for handling events in a worker environment.
    
    This class provides a structure for processing events with metadata tracking.
    Subclasses should implement the `handler` method to define specific event processing logic.
    
    Considerations: This interface is designed to be extended for various worker implementations, starting with Runpod.

    Attributes:
        context (dict): A dictionary to hold contextual information for the handler; this can be modified as needed.
        metadata (dict): A dictionary to track metadata about the handler's operations.
    """
    context: dict = field(default_factory=dict)
    metadata: dict = field(default_factory=dict)
    custom_actions: dict = field(default_factory=dict)

    def __post_init__(self):
        self.metadata["handler_created_at"] = datetime_utcnow().isoformat()

    @classmethod
    def find_handler(
            cls,
            import_pattern: str,
            handler_classname: str,
            **init_kwargs,
    ) -> 'HandlerInterface':
        import importlib

        # Dynamically import the handler class
        handler_module = importlib.import_module(import_pattern)
        handler_cls = getattr(handler_module, handler_classname)
        # Ensure the handler class exists and is a subclass of HandlerInterface
        if not handler_cls or not issubclass(handler_cls, cls):
            logger.error(f"Handler class '{handler_classname}' not found or is not a subclass of HandlerInterface: {handler_cls}")
            raise ValueError(f"Handler '{handler_classname}' not found in module '{import_pattern}' or is not a subclass of HandlerInterface.")
        kwargs = {
            "metadata": {
                "handler_found_at": datetime_utcnow().isoformat()
            },
            **init_kwargs,
        }
        return handler_cls.with_custom_actions(**kwargs)
    
    @classmethod
    def with_custom_actions(cls, actions: Optional[dict] = None, **init_kwargs) -> 'HandlerInterface':
        return cls(**init_kwargs).register_actions(actions=actions)

    def register_actions(self, actions: Optional[dict] = None) -> 'HandlerInterface':
        """Register multiple custom actions from a dictionary. You can chain this method after instantiation.
        Consider extending this method on child classes to register additinal custom actions:

        class MyHandler(HandlerInterface):
            
            def my_custom_action_method(self, **kwargs) -> dict:
                return {"status": "custom action executed", **kwargs}

            def register_actions(self, actions: Optional[dict] = None) -> 'HandlerInterface':
                # Call the parent method to register base actions
                super().register_actions(actions=actions)
                # Register additional custom actions specific to MyHandler
                self.register_custom_action(
                    action_name="my_custom_action",
                    action_callable=self.my_custom_action_method
                )
                return self

        Args:
            actions (dict): A dictionary where keys are action names and values are callables.
        Returns:
            HandlerInterface: The instance itself to allow method chaining.
        """
        # Register provided custom actions as arguments
        for action_name, action_callable in (actions or {}).items():
            self.register_custom_action(action_name=action_name, action_callable=action_callable)
        # Example custom action, just for demonstration purposes (e.g., a ping action)
        self.register_custom_action(
            action_name="ping",
            action_callable=lambda **kwargs: {
                "reply": "pong",
                "reply_at": datetime_utcnow().isoformat(),
                **kwargs,
            }
        )
        return self

    def register_custom_action(self, action_name: str, action_callable: Callable, ignore_if_exists: bool = False):
        """Register a custom action that can be invoked via the `fred_worker_action` key in the event payload.
        Args:
            action_name (str): The name of the custom action to register.
            action_callable (Callable): A callable (function or method) that implements the action.
            ignore_if_exists (bool): If True, will not overwrite an existing action with the same name.
        Raises:
            ValueError: If the action_callable is not callable.
        """
        if not callable(action_callable):
            raise ValueError(f"The action_callable must be a callable function or method: {type(action_callable)}")
        if action_name in self.custom_actions:
            if ignore_if_exists:
                logger.info(f"Custom action '{action_name}' already exists; ignoring as per flag.")
                return
            logger.warning(f"Overwriting existing custom action: '{action_name}'")
        self.custom_actions[action_name] = action_callable


    def telemetry(self, include_modules: bool = False) -> dict:
        from fred.utils.runtime import RuntimeInfo
        runtime_info = RuntimeInfo.auto()
        return runtime_info.to_dict(
            exclude_modules=not include_modules
        )

    def handler(self, payload: dict) -> Optional[dict]:
        logger.warning("Handler method not implemented.")
        return payload

    @property
    def metadata_prepared(self) -> dict:
        if not int(get_environ_variable("FRD_ENFORCE_METADATA_SERIALIZATION", default="0")):
            return self.metadata
        import json
        # Ensure serializability
        # TODO: Allow custom serialization methods
        metadata_serialized = json.dumps(self.metadata, default=str)
        return json.loads(metadata_serialized)

    @overload
    def run(self, event: dict, as_future: bool = True, future_id: Optional[str] = None) -> Future[dict]:
        ...

    @overload
    def run(self, event: dict, as_future: bool = False, future_id: Optional[str] = None) -> dict:
        ...

    def run(self, event: dict, as_future: bool = False, future_id: Optional[str] = None) -> dict | Future[dict]:
        """Process an incoming event and return a structured response.
        The event is expected to be a dictionary with at least an 'id' and 'input' keys.
        The 'input' key should contain the payload to be processed.
        Args:
            event (dict): The incoming event containing 'id' and 'input'.
            as_future (bool): If True, the processing will be done in a Future; otherwise, it will be synchronous.
        Returns:
            dict | Future[dict]: A structured response containing the result of processing the event.
                If requested as a Future, returns a Future that will resolve to the response dictionary.
        """
        if as_future:
            return Future(function=lambda: self.run(event=event, as_future=False), future_id=future_id)
        # Extract payload and event ID
        payload = event.get("input", {})
        job_event_identifier = event.get("id")
        # Update metadata for this run instance and timing information
        self.metadata["run_seq"] = self.metadata.get("run_seq", 0) + 1
        started_at = datetime_utcnow().isoformat()
        start_time = time.perf_counter()
        # Default response values
        ok = True
        response = None
        # Determine action based on 'fred_worker_action' in payload
        propagate_worker_error = int(payload.pop(
            "propagate_worker_error",
            get_environ_variable(
                "FRD_PROPAGATE_WORKER_ERROR",
                default="0"
            )
        ))
        match (worker_action := payload.pop("fred_worker_action", "handler")):
            case "telemetry":
                # Collect and return telemetry data
                logger.debug("Standard telemetry action requested.")
                response = self.telemetry()
            case "handler":
                # Process the payload using the handler method
                logger.debug("Standard handler action requested.")
                try:
                    response = self.handler(payload=payload)
                except Exception as e:
                    ok = False
                    logger.error(f"Error processing handler for event {job_event_identifier}: {e}")
                    if propagate_worker_error:
                        raise
                    response = {
                        "error": str(e)
                    }
            case action if isinstance(action, str):
                # Handle custom actions defined in the custom_actions dictionary
                logger.info(f"Custom fred_worker_action '{action}' received.")
                match self.custom_actions.get(action):
                    case None:
                        ok = False
                        response = {
                            "error": f"Custom action '{action}' not found."
                        }
                    case custom_action_function if callable(custom_action_function):
                        try:
                            response = custom_action_function(**payload)
                        except Exception as e:
                            ok = False
                            logger.error(f"Error processing custom action '{action}' for event {job_event_identifier}: {e}")
                            if propagate_worker_error:
                                raise
                            response = {
                                "error": str(e)
                            }
                    case _:
                        ok = False
                        logger.error(f"Custom action '{action}' is not callable.")
                        if propagate_worker_error:
                            raise ValueError(f"Custom action '{action}' is not callable.")
                        response = {
                            "error": f"Custom action '{action}' is not callable."
                        }
            case _:
                # Handle invalid action types
                logger.error(f"Invalid fred_worker_action type received: {type(worker_action)}")
                if propagate_worker_error:
                    raise ValueError("Invalid fred_worker_action type.")
                ok = False
                response = {
                    "error": "Invalid fred_worker_action type."
                }
        return {
            "ok": ok,
            "id": job_event_identifier,
            "started_at": started_at,
            "duration": time.perf_counter() - start_time,
            "worker_action": worker_action,
            "response": response,
            "metadata": self.metadata_prepared,
        }



import json
from time import perf_counter
from dataclasses import dataclass
from typing import (
    Callable,
    Generic,
    Optional,
    TypeVar,
)

from fred.settings import (
    get_environ_variable,
    logger_manager,
)
from fred.future.settings import (
    FRD_FUTURE_BACKEND,
    FRD_FUTURE_DEFAULT_EXPIRATION,
    FRD_FUTURE_DEFAULT_TIMEOUT,
)
from fred.future.callback.interface import CallbackInterface
from fred.dao.service.catalog import ServiceCatalog
from fred.utils.dateops import datetime_utcnow
from fred.dao.comp.catalog import FredKeyVal
from fred.monad.catalog import EitherMonad

logger = logger_manager.get_logger(__name__)

A = TypeVar("A")


class FutureBackend:
    keyval: type[FredKeyVal]

    @classmethod
    def with_backend(cls, service: ServiceCatalog, **kwargs) -> type['FutureBackend']:
        """Dynamically creates a FutureBackend subclass with the specified service backend.
        This method uses the service catalog to fetch the appropriate components for the
        given service and constructs a new subclass of FutureBackend with those components.
        Args:
            service (ServiceCatalog): The service catalog entry representing the backend service.
            **kwargs: Additional keyword arguments to pass to the component catalog.
        Returns:
            type[FutureBackend]: A new subclass of FutureBackend configured with the specified backend.
        """
        components = service.component_catalog(**kwargs)
        return type(
            f"{service.name.title()}{cls.__name__}",
            (cls,),
            {
                "keyval": components.KEYVAL.value
            },
        )

    @classmethod
    def infer_backend(cls, env: Optional[str] = None, **kwargs) -> type['FutureBackend']:
        """Infers the backend service from the environment variable or defaults to FRD_FUTURE_BACKEND.
        This method checks for an environment variable to determine the backend service. If the
        environment variable is not set, it defaults to the value of FRD_FUTURE_BACKEND. It then
        uses the inferred service to create a FutureBackend subclass with the appropriate components.
        Args:
            env (Optional[str]): The name of the environment variable to check for the backend service.
                                 If None, defaults to FRD_FUTURE_BACKEND.
            **kwargs: Additional keyword arguments to pass to the component catalog.
        Returns:
            type[FutureBackend]: A new subclass of FutureBackend configured with the inferred backend.
        """
        service_key = get_environ_variable(env, default=ServiceCatalog.STDLIB.name) \
            if env else FRD_FUTURE_BACKEND
        return cls.with_backend(
            service=ServiceCatalog[service_key.upper()],
            **kwargs
        )


@dataclass(frozen=True, slots=False)
class FutureResult(Generic[A], FutureBackend.infer_backend()):
    """Represents the result of an asynchronous computation (Future).
    This class provides methods to manage and retrieve the status and output of a Future.
    It uses a key-value store to persist the state and result of the Future, allowing for
    retrieval and management of asynchronous tasks."""
    future_id: str

    @staticmethod
    def _get_future_keyname(future_id: str) -> str:
        return ":".join(["frd", "future", future_id])

    @property
    def future_keyname(self) -> str:
        return self._get_future_keyname(future_id=self.future_id)
    
    @classmethod
    def _get_status_key(cls, future_id: str) -> FredKeyVal:
        return cls.keyval(key=":".join([cls._get_future_keyname(future_id=future_id), "status"]))

    @property
    def status(self) -> FredKeyVal:
        return self._get_status_key(future_id=self.future_id)
    
    @classmethod
    def _get_output_key(cls, future_id: str) -> FredKeyVal:
        return cls.keyval(key=":".join([cls._get_future_keyname(future_id=future_id), "output"]))

    @property
    def output(self) -> FredKeyVal:
        return self._get_output_key(future_id=self.future_id)
    
    @classmethod
    def _get_obj_key(cls, future_id: str) -> FredKeyVal:
        return cls.keyval(key=":".join([cls._get_future_keyname(future_id=future_id), "obj"]))

    @property
    def obj(self) -> FredKeyVal:
        return self._get_obj_key(future_id=self.future_id)

    def stringify(self) -> str:
        import pickle
        import base64
        return base64.b64encode(pickle.dumps(self)).decode("ascii")

    @classmethod
    def from_string(cls, payload: str) -> 'FutureResult[A]':
        import pickle
        import base64
        return pickle.loads(base64.b64decode(payload))
    
    def _from_backend(self) -> Optional['FutureResult[A]']:
        payload = self.obj.get()
        if not payload:
            return
        return self.from_string(payload=payload)

    @classmethod
    def from_backend(cls, future_id: str) -> Optional['FutureResult[A]']:
        return FutureResult(future_id=future_id)._from_backend()


@dataclass(frozen=True, slots=False)
class FutureUndefinedPending(FutureResult[A]):

    @classmethod
    def auto(cls, **kwargs) -> 'FutureUndefinedPending[A]':
        """Creates a FutureUndefinedPending instance with an auto-generated future_id.
        This method generates a unique identifier for the future_id if one is not provided
        in the keyword arguments. It ensures that each instance has a distinct future_id,
        which is essential for tracking and managing asynchronous tasks.
        Args:
            **kwargs: Additional keyword arguments to pass to the FutureUndefinedPending constructor.
        Returns:
            FutureUndefinedPending[A]: A new instance of FutureUndefinedPending with a unique future_id.
        """
        import uuid
        future_id = kwargs.pop("future_id", None) or str(uuid.uuid4())
        return FutureUndefinedPending[A](future_id=future_id, **kwargs)

    def __post_init__(self):
        logger.debug(f"Future[{self.future_id}] initialized and pending execution")
        self.status.set(
            value=f"UNDEFINED:PENDING:{datetime_utcnow().isoformat()}",
            expire=FRD_FUTURE_DEFAULT_EXPIRATION
        )
        self.obj.set(
            value=self.stringify(),
            expire=FRD_FUTURE_DEFAULT_EXPIRATION
        )

    def apply(
            self,
            function: Callable[..., A],
            on_start: Optional[CallbackInterface] = None,
            on_complete: Optional[CallbackInterface] = None,
            **kwargs
        ) -> 'FutureDefined[A]':
        """Applies a function to the Future, transitioning it from pending to in-progress and finalizing as defined.
        This method executes the provided function with the given keyword arguments,
        transitioning the Future from a pending state to an in-progress state. It captures
        the result of the function execution, which can be either a successful value or an
        exception, and returns a FutureDefined instance representing the completed state.
        Args:
            function (Callable[..., A]): The function to execute, which should return a value of type A.
            **kwargs: Additional keyword arguments to pass to the function during execution.
        Returns:
            FutureDefined[A]: A new instance of FutureDefined representing the completed state of the Future.
        """
        fip = FutureUndefinedInProgress[A](
            future_id=self.future_id,
            started_at=perf_counter(),
            function_name=function.__name__,
        )
        return fip.exec(
            function=function,
            on_start=on_start,
            on_complete=on_complete,
            **kwargs
        )


@dataclass(frozen=True, slots=False)
class FutureUndefinedInProgress(FutureResult[A]):
    started_at: float
    function_name: str
    

    def __post_init__(self):
        logger.debug(f"Future[{self.future_id}] started execution of function '{self.function_name}'")
        self.status.set(
            value=f"UNDEFINED:IN_PROGRESS:{datetime_utcnow().isoformat()}",
            expire=FRD_FUTURE_DEFAULT_EXPIRATION
        )
        self.obj.set(
            value=self.stringify(),
            expire=FRD_FUTURE_DEFAULT_EXPIRATION
        )

    def exec(
            self,
            function: Callable[..., A],
            on_start: Optional[CallbackInterface] = None,
            on_complete: Optional[CallbackInterface] = None,
            fail: bool = False,
            **kwargs,
        ) -> 'FutureDefined[A]':
        """Executes the function associated with the Future, capturing its result or exception.
        This method runs the provided function with the given keyword arguments, capturing
        its output. If the function executes successfully, it wraps the result in a Right
        monad; if it raises an exception, it captures the exception in a Left monad.
        The method returns a FutureDefined instance containing the result of the execution.
        If the 'fail' parameter is set to True, any exception raised during function execution
        will be propagated instead of being captured in the Left monad.
        Args:
            function (Callable[..., A]): The function to execute, which should return a value of type A.
            fail (bool): If True, exceptions raised during function execution will be propagated. Defaults to False.
            **kwargs: Additional keyword arguments to pass to the function during execution.
        Returns:
            FutureDefined[A]: A new instance of FutureDefined representing the completed state of the Future.
        """
        # Execute on_start callback if provided
        on_start_thread = (
            logger.debug(f"Future[{self.future_id}] executing on_start callback")
            or on_start.run(future_id=self.future_id, blocking=False)
        ) if on_start else None
        try:
            ok = False
            match function(**kwargs):
                case EitherMonad.Right(value=value):
                    ok = True
                    value =EitherMonad.Right.from_value(val=value)
                case EitherMonad.Left(exception=exception):
                    ok = False
                    value = EitherMonad.Left.from_value(val=exception)
                case value:
                    ok = True
                    value = EitherMonad.Right.from_value(val=value)      
        except Exception as e:
            ok = False
            value = EitherMonad.Left.from_value(val=e)
            if fail:
                raise e
        future_defined = FutureDefined(
            future_id=self.future_id,
            value=value,
            ok=ok,
        )
        # We could do the following to have an "on_success" behaviour:
        # value.map(lambda v: on_complete.run(v) if on_complete else None)
        # But that would leave out the "on_failure" behaviour and we would probably need to
        # add specific 'on_success' and 'on_failure' callbacks to mitigate that.
        # Instead, we just do a generic 'on_complete' callback that gets executed
        # wrapped on an Either monad so the user can handle success/failure as needed.
        if on_complete:
            logger.debug(f"Future[{self.future_id}] executing on_complete callback")
            on_complete.run(future_id=self.future_id, output=value, blocking=True)
        if on_start_thread:
            on_start_thread.join()
        return future_defined


@dataclass(frozen=True, slots=False)
class FutureDefined(FutureResult[A]):
    value: EitherMonad.Either[A]
    ok: bool

    def __post_init__(self):
        state = "SUCCESS" if self.ok else "FAILURE"
        logger.debug(f"FutureDefined[{self.future_id}] completed with state: {state}")
        self.status.set(
            value=f"DEFINED:{state}:{datetime_utcnow().isoformat()}",
            expire=FRD_FUTURE_DEFAULT_EXPIRATION
        )
        self.obj.set(
            value=self.stringify(),
            expire=FRD_FUTURE_DEFAULT_EXPIRATION
        )
        match self.value:
            case EitherMonad.Right(value=value):
                # TODO: Consider using a more robust serialization method...
                # TODO: Should we consider collapsing nested-futures? i.e., Future[Future[A]] => Future[A]
                payload = json.dumps(value)
                self.output.set(
                    value=payload,
                    expire=FRD_FUTURE_DEFAULT_EXPIRATION
                )
            case EitherMonad.Left(exception=exception):
                payload = json.dumps({"error": str(exception)})
                self.output.set(
                    value=str(exception),
                    expire=FRD_FUTURE_DEFAULT_EXPIRATION
                )



import uuid
from dataclasses import dataclass

from fred.settings import logger_manager
from fred.utils.dateops import datetime_utcnow
from fred.worker.runner.model._handler import Handler
from fred.worker.runner.model.interface import ModelInterface
from fred.worker.runner.settings import (
    FRD_RUNNER_REQUEST_QUEUE,
    FRD_RUNNER_RESPONSE_QUEUE,
)

logger = logger_manager.get_logger(name=__name__)



@dataclass(frozen=True, slots=False)
class RunnerSpec(ModelInterface):
    runner_id: str
    created_at: str
    queue_slug: str
    inner: Handler
    use_response_queue: bool = False
    lifetime: int = 3600  # Default to 1 hour if not specified
    timeout: int = 30  # Default to 30 seconds if not specified

    @classmethod
    def auto(cls, **kwargs) -> "RunnerSpec":
        kwargs["runner_id"] = kwargs.get("runner_id", str(uuid.uuid4()))
        return cls.from_payload(payload=kwargs)

    @classmethod
    def from_payload(cls, payload: dict) -> "RunnerSpec":
        inner_handler_kwargs = {
            "classname": payload.pop("handler_classname", None) or (
                logger.warning("No handler_classname provided; defaulting to: 'HandlerInterface'")
                or "HandlerInterface"
            ),
            "classpath": payload.pop("handler_classpath", None) or (
                logger.warning("No handler_classpath provided; defaulting to: 'fred.worker.interface'")
                or "fred.worker.interface"
            ),
            "kwargs": payload.pop("handler_kwargs", {}),
            **payload.pop("handler_configs", {})
        }
        runner_id = payload.pop("runner_id", None) or (
            logger.warning("No runner_id provided; generating a new one using UUID4.")
            or str(uuid.uuid4())
        )
        return cls(
            runner_id=runner_id,
            created_at=datetime_utcnow().isoformat(),
            queue_slug=payload.pop("queue_slug", None) or (
                logger.warning("Queue slug not specified; defaulting to: 'demo'")
                or "demo"
            ),
            inner=Handler(**inner_handler_kwargs),
            use_response_queue=payload.pop("use_response_queue", False),
            lifetime=payload.pop("lifetime", 3600),
            timeout=payload.pop("timeout", 30),
        )
    
    def as_dict(self) -> dict:
        return {
            "runner_id": self.runner_id,
            "created_at": self.created_at,
            "queue_slug": self.queue_slug,
            "handler_classname": self.inner.classname,
            "handler_classpath": self.inner.classpath,
            "handler_kwargs": self.inner.kwargs,
            "use_response_queue": self.use_response_queue,
            "lifetime": self.lifetime,
            "timeout": self.timeout,
        }
    
    def as_event(self) -> dict:
        return {
            "id": self.runner_id,
            "input": self.as_dict(),
        }
    
    @property
    def request_queue_name(self) -> str:
        return FRD_RUNNER_REQUEST_QUEUE or f"req:{self.queue_slug}"
    
    @property
    def response_queue_name(self) -> str:
        return FRD_RUNNER_RESPONSE_QUEUE or f"res:{self.queue_slug}"



import enum

from fred.worker.runner.rest.routers.interface import RouterInterface
from fred.worker.runner.rest.routers._runner import RunnerRouter


class RouterCatalog(enum.Enum):
    EXAMPLE = RouterInterface.auto()  # Default example router directly from the base interface
    RUNNER = RunnerRouter.auto(
        prefix="/runner",
        tags=["Runner"],
    )

    def get_router_instance(self) -> RouterInterface:
        return self.value.router

    def get_router_configs(self) -> dict:
        return self.value.router_configs



from typing import Optional

from fred.cli.interface import IntegrationExtCLI
from fred.settings import logger_manager


logger = logger_manager.get_logger(name=__name__)


class RunnerServerExt(IntegrationExtCLI):

    def start(
            self,
            include_routers: Optional[list[str]] = None,
            exclude_routers: Optional[list[str]] = None,
            fast_api_configs: Optional[dict] = None,
            server_configs: Optional[dict] = None,
        ):
        from fred.worker.runner.rest.server import RunnerServer

        include_routers = include_routers or []
        exclude_routers = exclude_routers or []
        fast_api_configs = fast_api_configs or {}
        server_configs = server_configs or {}

        logger.info("Starting REST server...")
        server = RunnerServer.auto(
            include_routers=include_routers,
            exclude_routers=exclude_routers,
            **fast_api_configs,
        )
        server.start(**server_configs)



from typing import Optional

from fred.cli.interface import IntegrationExtCLI
from fred.settings import logger_manager


logger = logger_manager.get_logger(name=__name__)


class DatabricksExt(IntegrationExtCLI):
    def sync_runtime(self, runtime: Optional[str] = None):
        from fred.integrations.databricks.runtimes.sync import DatabricksRuntimeSyncHelper

        sync_helper = DatabricksRuntimeSyncHelper.default()
        match runtime:
            case None:
                logger.info("No runtime specified; skipping synchronization.")
            case "all":
                logger.info("Syncing all available Databricks runtimes.")
                sync_helper.sync_all()
            case runtime_name if isinstance(runtime_name, str):
                logger.info(f"Syncing specified Databricks runtime: {runtime_name}")
                sync_helper.sync(runtime=runtime_name)
            case _:
                raise ValueError(f"Invalid runtime value: {runtime}")



import datetime as dt
from dataclasses import dataclass, asdict

from fred.utils.dateops import datetime_utcnow


@dataclass(frozen=True, slots=True)
class RuntimeProfilingSnapshot:
    snapshot_at: dt.datetime
    cpu_percent: float
    virtual_memory_percent: float
    swap_memory_percent: float
    disk_usage_percent: float

    @classmethod
    def auto(cls) -> 'RuntimeProfilingSnapshot':
        import psutil

        return cls(
            snapshot_at=datetime_utcnow(),
            cpu_percent=psutil.cpu_percent(interval=1),
            virtual_memory_percent=psutil.virtual_memory().percent,
            swap_memory_percent=psutil.swap_memory().percent,
            disk_usage_percent=psutil.disk_usage("/").percent,
        )
    
    def to_dict(self) -> dict:
        payload = asdict(self)
        payload["snapshot_at"] = self.snapshot_at.isoformat()
        return payload


@dataclass(frozen=True, slots=True)
class RuntimeInfo:
    snapshot_at: dt.datetime
    python_version: str
    platform: str
    processor: str
    modules: list[str]
    profiling_snapshots: list[RuntimeProfilingSnapshot]


    @classmethod
    def auto(cls, exclude_initial_profile: bool = False) -> 'RuntimeInfo':
        import platform

        return cls(
            python_version=platform.python_version(),
            platform=platform.platform(),
            processor=platform.processor(),
            modules=sorted(cls.get_modules()),
            snapshot_at=datetime_utcnow(),
            profiling_snapshots=[] if exclude_initial_profile else [
                RuntimeProfilingSnapshot.auto()
            ],
        )
    
    def append_profiling_snapshot(self):
        self.profiling_snapshots.append(
            RuntimeProfilingSnapshot.auto()
        )

    @staticmethod
    def get_modules():
        import sys
        import pkgutil

        return {
            module.name
            for module in pkgutil.iter_modules()
        }.union(sys.builtin_module_names)
    
    def to_dict(
            self,
            exclude_modules: bool = False,
            exclude_profiling_snapshots: bool = False,
        ) -> dict:
        payload = asdict(self)
        payload["snapshot_at"] = self.snapshot_at.isoformat()
        payload["profiling_snapshots"] = [
            snapshot.to_dict() for snapshot in self.profiling_snapshots
        ]
        if exclude_modules:
            payload.pop("modules", None)
        if exclude_profiling_snapshots:
            payload.pop("profiling_snapshots", None)
        return payload


from enum import Enum, auto

from fred.dao.comp.catalog import FredQueue


class RunnerSignal(Enum):
    """Signals used for controlling the runner process."""
    STOP = auto()
    PING = auto()

    def send(self, queue: FredQueue):
        """Send the signal to the specified queue."""
        return queue.add(item=self.name)



from fred.maturity import Maturity, MaturityLevel


module_maturity = Maturity(
    level=MaturityLevel.ALPHA,
    reference=__name__,
    message=(
        "Databricks integration is in early development "
        "and therefore currently with incomplete and unstable features."
    )
)


import time
from typing import TypeVar, Optional

from fred.future.settings import FRD_FUTURE_DEFAULT_EXPIRATION

A = TypeVar("A")


def pull_future_result(
        future_id: str,
        retry_delay: float = 0.2,
        retry_backoff_rate: float = 0.1,
        retry_delay_max: float = 15,
        timeout: float = FRD_FUTURE_DEFAULT_EXPIRATION,
    ) -> A:
    from fred.future.impl import FutureResult, FutureUndefinedInProgress, FutureUndefinedPending

    if timeout <= 0:
        raise ValueError(f"Pull operation for future_id '{future_id}' has timed out.")

    match FutureResult.from_backend(future_id=future_id):
        case None:
            raise ValueError(f"No future found with ID '{future_id}'")
        case FutureResult(value=value):
            return value.resolve()
        case FutureUndefinedPending() | FutureUndefinedInProgress():
            # If the future is not yet defined, wait for it to complete
            time.sleep(retry_delay)
            # Increase the delay for the next check to avoid busy waiting (exponential backoff) capped at delay_max
            return pull_future_result(
                future_id=future_id,
                retry_delay=min(retry_delay * (1 + retry_backoff_rate), retry_delay_max),
                retry_backoff_rate=retry_backoff_rate,
                retry_delay_max=retry_delay_max,
                # TODO: Consider using a more precise timeout mechanism based on elapsed time
                timeout=timeout - retry_delay,
            )
        case _:
            raise ValueError(f"Unknown future state for ID '{future_id}'")



import enum
from dataclasses import dataclass
from typing import Optional


from fred.settings import (
    logger_manager,
    get_environ_variable,
)


logger = logger_manager.get_logger(name=__name__)


class MaturityLevel(enum.Enum):
    ALPHA = (
        "Alpha Maturity Level"
        ": Initial development stage with potential instability and limited or unreliable features."
    )
    BETA = (
        "Beta Maturity Level"
        ": More stable than Alpha, with additional features and ongoing testing."
    )
    STABLE = (
        "Stable Maturity Level"
        ": Fully tested and mostly reliable, suitable for production use, but use at your own risk!"
    )
    TO_BE_DEPRECATED = (
        "To Be Deprecated Maturity Level"
        ": Functionality is planned to be removed in future releases. Consider alternatives."
    )
    DEPRECATED = (
        "Deprecated Maturity Level"
        ": Functionality is no longer supported and may be removed in future releases."
    )
    REMOVED = (
        "Removed Maturity Level"
        ": Functionality has been removed from the codebase and is no longer available."
    )

    def is_stable(self) -> bool:
        return self == MaturityLevel.STABLE
    
    def is_unstable(self) -> bool:
        return self in {MaturityLevel.ALPHA, MaturityLevel.BETA, MaturityLevel.TO_BE_DEPRECATED}
    
    def is_deprecated(self) -> bool:
        return self in {MaturityLevel.DEPRECATED, MaturityLevel.REMOVED}


@dataclass(frozen=True, slots=True)
class Maturity:
    level: MaturityLevel
    reference: Optional[str] = None
    message: Optional[str] = None


    def __post_init__(self):
        if not self.level.is_stable() and not self.quiet:
            logger.warning(
                "Functionality for {reference} is at {level} maturity level. {message}".format(
                    reference=self.reference or "undefined",
                    level=self.level.name,
                    message=self.message or ""
                )
            )

    @property
    def quiet(self) -> bool:
        return get_environ_variable(
            name="FRD_DISABLE_MATURITY_WARN",
            default="0",
        ).upper() in ["1", "Y", "T", "TRUE", "YES"]



import uuid
import json
from dataclasses import dataclass
from typing import Optional

from fred.future import Future
from fred.monad.catalog import EitherMonad
from fred.utils.dateops import datetime_utcnow
from fred.dao.comp.catalog import FredQueue
from fred.worker.runner.settings import FRD_RUNNER_BACKEND
from fred.worker.runner.backend import RunnerBackend
from fred.worker.runner.model.catalog import RunnerModelCatalog
from fred.worker.interface import HandlerInterface
from fred.worker.runner.status import RunnerStatus

from fred.settings import logger_manager


logger = logger_manager.get_logger(name=__name__)


@dataclass(frozen=True, slots=False)
class RunnerHandler(HandlerInterface):
    
    def __post_init__(self):
        super().__post_init__()
        logger.info("Fred-Runner outer-handler initialized using Fred-Worker interface.")

    @staticmethod 
    def _runner_process(
            item: dict,
            runner: HandlerInterface,
            item_id: str,
            request_id: str,
    ) -> Future[dict]:
        logger.info(f"Processing item '{item_id}' provided by request-id: {request_id}")
        return runner.run(
            event={
                "id": item_id,
                "input": item
            },
            as_future=True,
            future_id=request_id,
        )

    def _runner_loop(
            self,
            runner: HandlerInterface,
            req_queue: FredQueue,
            lifespan: int,
            timeout: int,
            res_queue: Optional[FredQueue] = None,
    ) -> dict:
        start_time = datetime_utcnow()
        last_processed_time = datetime_utcnow()
        while (elapsed_seconds := (datetime_utcnow() - start_time).total_seconds()):
            if elapsed_seconds > lifespan:
                logger.info("Lifespan exceeded; exiting runner loop.")
                break
            if (idle_seconds := (datetime_utcnow() - last_processed_time).total_seconds()) > timeout:
                logger.info(f"Idle time ({idle_seconds}) exceeded timeout ({timeout}); exiting runner loop.")
                break
            # Fetch item from Redis queue 
            try:
                message = req_queue.pop()
                # If no item, iterate again
                if not message:
                    continue
            except Exception as e:
                logger.error(f"Error fetching item from Redis queue '{req_queue}': {e}")
                continue
            # Handle special signals
            match message:
                case "STOP" | "SHUTDOWN" | "TERMINATE":
                    logger.info("Received STOP signal; exiting runner loop.")
                    break
                case "PING":
                    logger.info("Received PING signal; continuing.")
                    last_processed_time = datetime_utcnow()
                    continue
                case _:
                    pass
            try:
                item = json.loads(message)
                item_id = item.get("item_id") or (
                    logger.warning("No item_id provided in item-payload; generating a new one using UUID5 hash.")
                    or str(uuid.uuid5(uuid.NAMESPACE_OID, message))
                )
                # The request_id is used as the future_id for tracking purposes
                request_id = item.get("request_id") or (
                    logger.warning(f"No request_id provided in item-payload; using the item_id '{item_id}' instead.")
                    or item_id
                )
            except Exception as e:
                logger.error(f"Error decoding or parsing item from Redis: {e}")
                continue
            # Process item using runner and handle result
            future = self._runner_process(item=item, runner=runner, item_id=item_id, request_id=request_id).map(
                lambda res: res if isinstance(res, str) else json.dumps(res, default=str)
            )
            match future.wait():
                case EitherMonad.Right(value):
                    if res_queue:
                        logger.debug(f"Processed item ID '{item_id}' on request ID '{request_id}' and pushed result to response queue.")
                        res_queue.add(value)
                case EitherMonad.Left(error):
                    logger.error(f"Error processing item ID '{item_id}' on request ID '{request_id}': {error}")
                    continue
                case _:
                    logger.error(f"Unexpected result processing item ID '{item_id}' on request ID '{request_id}': {future}")
                    continue
            last_processed_time = datetime_utcnow()
        return {
            "started_at": start_time.isoformat(),
            "stopped_at": datetime_utcnow().isoformat(),
            "total_elapsed_seconds": (datetime_utcnow() - start_time).total_seconds(),
            "last_processed_at": last_processed_time.isoformat(),
            "idle_seconds": (datetime_utcnow() - last_processed_time).total_seconds(),
        }
 

    def handler(self, payload: dict) -> dict:
        # Configure the backend service abstraction (e.g., Redis)
        runner_backend = RunnerBackend.auto(
            service_name=payload.pop("runner_backend", None) or FRD_RUNNER_BACKEND,
            **payload.pop("runner_backend_configs", {}),
        )
        # Outer handler model instance
        spec = RunnerModelCatalog.RUNNER_SPEC.value.from_payload(payload=payload)
        
        # Determine request and response queues to use for this runner instance
        req_queue = runner_backend.queue(name=spec.request_queue_name)
        res_queue = runner_backend.queue(name=spec.response_queue_name) \
            if spec.use_response_queue else None
        # Get runner (inner handler) instance and ID
        # The 'outer' handler delegates the processing of tasks to the 'inner' handler
        # and is responsible for managing the lifecycle and updating the status of the runner
        runner = spec.inner.get_handler()
        runner_id = spec.runner_id
        runner_status = runner_backend.keyval(
            key=RunnerStatus.get_key(runner_id=runner_id)
        )
        logger.info(f"Starting runner with ID '{runner_id}' using request-queue '{req_queue.name}'")
        runner_status.set(
            value=RunnerStatus.STARTED.get_val(),
            expire=None,
        )
        runner_loop = Future(
            function=self._runner_loop,
            runner=runner,
            req_queue=req_queue,
            lifespan=spec.lifetime,
            timeout=spec.timeout,
            res_queue=res_queue,
            # The runner_id is used as the future_id for tracking purposes
            future_id=runner_id,
        )
        runner_status.set(
            value=RunnerStatus.RUNNING.get_val(),
            expire=None,
        )
        results = {
            "runner_id": runner_id,
            "runner_started_at": datetime_utcnow().isoformat(),
            "runner_pending_requests": req_queue.size(),
        }
        match runner_loop.wait():
            case EitherMonad.Right(meta):
                results["runner_status"] = "success"
                results["metadata"] = meta
            case EitherMonad.Left(error):
                results["runner_status"] = "failure"
                results["metadata"] = {
                    "error": str(error)
                }
            case _:
                results["runner_status"] = "unexpected"
                results["metadata"] = {
                    "error": "Unexpected result from runner loop"
                }
        results["pending_requests"] = pending_requests = req_queue.size()
        runner_status.set(
            value=RunnerStatus.STOPPED.get_val(str(pending_requests)),
            expire=3600,  # Keep the stopped status for 1 hour
        )
        if pending_requests:
            logger.warning(f"Runner '{runner_id}' stopped with {pending_requests} pending items still in the queue: '{req_queue.name}'")
        else:
            logger.info(f"Runner '{runner_id}' stopped with no pending items in the queue: '{req_queue.name}'")

        return results



import enum
from functools import lru_cache
from typing import TypeVar, Optional

from fred.dao.service.interface import ServiceInterface
from fred.dao.service._redis import RedisService
from fred.dao.service._stdlib import StdLibService

T = TypeVar("T")


class ServiceCatalog(enum.Enum):
    STDLIB = StdLibService
    REDIS = RedisService

    @classmethod
    def from_classname(cls, classname: str) -> "ServiceCatalog":
        for item in cls:
            if item.value.__name__ == classname:
                return item
        raise ValueError(f"No service found for classname: {classname}")

    @lru_cache(maxsize=None)  # TODO: Consider cache invalidation strategy if needed
    def component_catalog(self, srv_ref: Optional[str | ServiceInterface] = None, **kwargs) -> enum.Enum:
        """Get a preconfigured component catalog for this (self) service.
        This method returns a new Enum with preconfigured components for the
        service represented by this enum member.
        Args:
            **kwargs: Additional keyword arguments to pass to the component constructors.
        Returns:
             enum.Enum: A new Enum with preconfigured components for this service.
        """
        from fred.dao.comp.catalog import CompCatalog  # Avoid circular import

        return CompCatalog.preconf(srv_ref=srv_ref or self.name, **kwargs)

    def service_cls(self) -> type[ServiceInterface]:
        return self.value

    def auto(self, **kwargs) -> ServiceInterface[T]:
        return self.value.auto(**kwargs)



import time
from dataclasses import dataclass
from typing import Optional

from fred.future import Future
from fred.future.result import FutureResult
from fred.dao.comp.catalog import FredQueue
from fred.worker.runner.status import RunnerStatus
from fred.worker.runner.signal import RunnerSignal
from fred.worker.runner.backend import RunnerBackend
from fred.worker.runner.model.catalog import RunnerModelCatalog
from fred.worker.runner.settings import (
    FRD_RUNNER_BACKEND,
    FRD_RUNNER_REQUEST_QUEUE,
    FRD_RUNNER_RESPONSE_QUEUE,
)
from fred.settings import logger_manager

logger = logger_manager.get_logger(name=__name__)


@dataclass(frozen=True, slots=True)
class RunnerClient:
    _runner_backend: RunnerBackend
    req_queue: FredQueue
    res_queue: FredQueue


    @classmethod
    def auto(
            cls,
            queue_slug: Optional[str] = None,
            service_name: Optional[str] = None,
            **kwargs
        ) -> "RunnerClient":
        queue_slug = queue_slug or kwargs.pop("queue_slug", None) or (
            logger.warning("Queue slug not specified; defaulting to: 'demo'")
            or "demo"
        )
        queue_name_request = FRD_RUNNER_REQUEST_QUEUE or f"req:{queue_slug}"
        queue_name_response = FRD_RUNNER_RESPONSE_QUEUE or f"res:{queue_slug}"
        runner_backend = RunnerBackend.auto(
            service_name=service_name or FRD_RUNNER_BACKEND,
            **kwargs
        )
        return cls(
            _runner_backend=runner_backend,
            req_queue=runner_backend.queue(name=queue_name_request),
            res_queue=runner_backend.queue(name=queue_name_response),
        )

    @property
    def PING(self):
        return self.signal(RunnerSignal.PING)

    @property
    def STOP(self):
        return self.signal(RunnerSignal.STOP)

    def signal(self, signal: str | RunnerSignal):
        signal = RunnerSignal[signal.upper()] if isinstance(signal, str) else signal
        return signal.send(self.req_queue)

    def runner_status(self, runner_id: str) -> RunnerStatus:
        runner_status=self._runner_backend.keyval(
                key=RunnerStatus.get_key(runner_id=runner_id)
            )
        if not (value := runner_status.get()):
            logger.warning(f"No status found for runner_id: '{runner_id}'")
            return RunnerStatus.UNDEFINED
        return RunnerStatus.parse_value(value=value)

    def send(
            self,
            item: dict,
            req_uuid_hash: bool = False,
            item_uuid_hash: bool = False,
    ) -> str:
        item_instance = RunnerModelCatalog.ITEM.value.uuid(payload=item, uuid_hash=item_uuid_hash)
        request = item_instance.as_request(
            use_hash=req_uuid_hash,
            request_id=item.get("request_id"),
        )
        request.dispatch(request_queue=self.req_queue)
        return request.request_id
    
    @staticmethod
    def fetch_status(request_id: str) -> Optional[str]:
        return FutureResult._get_status_key(future_id=request_id).get()

    def pullsync(
            self,
            request_id: str,
            retry_sync: int = 10,
            retry_delay: float = 0.2,
            retry_backoff_rate: float = 0.1,
            **kwargs,
    ) -> Future:
        # TODO: Once the broadcast method is implemented, we can add a warning notice regarding
        # the pullsync metho adding extra load to the backend service
        future = Future(
            function=self._is_ready_for_pullsync,
            request_id=request_id,
            retry_sync=retry_sync,
            retry_delay=retry_delay,
            retry_backoff_rate=retry_backoff_rate,
            fail=True,  # Always fail inside the future to propagate the exception
        )
        return future.flat_map(
            lambda _: Future.pullsync(future_id=request_id, **kwargs)
        )

    def _is_ready_for_pullsync(
            self,
            request_id: str,
            retry_sync: int = 10,
            retry_delay: float = 0.2,
            retry_backoff_rate: float = 0.1,
            fail: bool = False,
    ) -> bool:
        if not self.fetch_status(request_id=request_id):
            logger.info(f"No status found for request_id '{request_id}'")
            retry_kwargs = {
                "request_id": request_id,
                "retry_sync": retry_sync - 1,
                "retry_delay": retry_delay * (1 + retry_backoff_rate),  # Exponential backoff
                "retry_backoff_rate": retry_backoff_rate,
                "fail": fail,
            }
            if retry_sync:
                time.sleep(retry_delay)
                return self._is_ready_for_pullsync(**retry_kwargs)
            elif fail:
                logger.error(f"Failed to fetch status for request_id '{request_id}' after retries.")
                raise ValueError(f"No status found for request_id '{request_id}'")
            else:
                return False
        return True

    def fetch_result(self, request_id: str, now: bool = False, timeout: Optional[float] = None) -> Optional[dict]:
        future = self.pullsync(request_id=request_id)
        if now:
            return future.getwhatevernow()
        return future.wait_and_resolve(timeout=timeout)



import os
import json
from typing import Dict, List
from setuptools import setup, find_namespace_packages


CODEBASE_PATH = os.environ.get(
    "CODEBASE_PATH",
    default=os.path.join("src", "main"),
)

with open("requirements.txt", "r") as file:
    requirements = [line for line in file.read().splitlines() if line and not line.startswith("#")]

version_filepath = os.path.join(CODEBASE_PATH, "fred", "version")
with open(version_filepath, "r") as file:
    version = file.read().strip()

with open("README.md") as file:
    readme = file.read()

setup(
    name="fred-oss",
    version=version,
    description="FREDOSS",
    long_description=readme,
    long_description_content_type='text/markdown',
    url="https://fred.fahera.mx",
    author="Fahera Research, Education, and Development",
    author_email="fred@fahera.mx",
    packages=find_namespace_packages(where=CODEBASE_PATH),
    package_dir={
        "": CODEBASE_PATH
    },
    package_data={
        "": [
            version_filepath,
        ]
    },
    entry_points={
        "console_scripts": [
            "fred=fred.cli.main:CLI.cli_exec",
        ]
    },
    install_requires=requirements,
    include_package_data=True,
    # Python version should be aligned to the latest databricks LTS runtime at the moment.
    # For more info: https://docs.databricks.com/aws/en/release-notes/runtime/ 
    # Reference: Databricks Runtime 16.4 LTS Using Python 3.12.3
    python_requires=">=3.12",
)



from fred.settings import get_environ_variable


FRD_FUTURE_BACKEND = get_environ_variable(
    "FRD_FUTURE_BACKEND",
    default="STDLIB",
)

FRD_FUTURE_DEFAULT_EXPIRATION = int(get_environ_variable(
    "FRD_FUTURE_DEFAULT_EXPIRATION",
    default="3600",  # 1 hour
))

# TODO: This is currently not used, but reserved for future use...
FRD_FUTURE_DEFAULT_TIMEOUT = int(get_environ_variable(
    "FRD_FUTURE_DEFAULT_TIMEOUT",
    default="900",  # 15 minutes
))



import os
import json
from dataclasses import dataclass
from typing import Union

from fred.integrations.databricks.runtime import DatabricksRuntimeCatalog
from fred.settings import logger_manager


logger = logger_manager.get_logger(name=__name__)

RUNTIME_DATATYPE = Union[DatabricksRuntimeCatalog, str]


@dataclass(frozen=True, slots=True)
class DatabricksRuntimeSyncHelper:
    output_path: str

    @classmethod
    def default(cls) -> 'DatabricksRuntimeSyncHelper':
        return cls(
            output_path=os.path.join(
                os.path.dirname(__file__),
                "runtimes",
            )
        )

    def get_payload(self, runtime: DatabricksRuntimeCatalog) -> list[dict]:
        import requests

        # Get the runtime content from the official documentation
        response = requests.get(runtime.url)
        if not response.ok:
            raise ValueError(f"Failed to fetch runtime information from {runtime.url}")
        content = response.text

        # Parse the content to extract relevant information
        # TODO: Implement the actual parsing logic
        raise NotImplementedError

    def sync(self, runtime: RUNTIME_DATATYPE):
        if isinstance(runtime, str):
            runtime = DatabricksRuntimeCatalog.from_string(runtime)
        logger.info(f"Syncing Databricks Runtime: {runtime.name}")
        payload = self.get_payload(runtime=runtime)
        output_filepath = os.path.join(
            self.output_path,
            f"{runtime.name.replace(' ', '')}.json",
        )
        with open(output_filepath, "w") as file:
            json.dump(payload, file, indent=4)

    def sync_all(self):
        for runtime in DatabricksRuntimeCatalog:
            self.sync(runtime=runtime)
